# pylint: disable=cyclic-import

import re as _re
import warnings as _warnings
from collections import OrderedDict as _OrderedDict
from copy import copy as _copy
from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List
from typing import Mapping as _Mapping

import torch.nn as _nn

from tamm._adapters_v1.adapted_layer import AdaptedLayer as _AdaptedLayer
from tamm._adapters_v1.layer_adapters.common import LayerAdapter as _LayerAdapter
from tamm._adapters_v1.layer_adapters.lora import LoRA as _LoRA
from tamm._adapters_v1.layer_adapters.lora import (
    LoRAFusedMultiOutputLinear as _LoRAFusedMultiOutputLinear,
)
from tamm.runtime_configuration import rc as _rc
from tamm.typing import StateDictType as _StateDictType
from tamm.utils import torch_utils as _torch_utils

_TORCH_CHECKPOINT_ACTIVATION_KEY = "._checkpoint_wrapped_module"


def is_v0_lora_state_dict(state_dict: _Mapping[str, _Any]) -> bool:
    """
    Returns True if the state_dict corresponds to a Tamm v0 state dict.
    """
    pattern = _re.compile(r"\.adapter\.(lora_W(q|k|v)\.)?[ab]_transpose$")
    any_adapter_key = any("adapter" in key for key in state_dict)
    if not any_adapter_key:
        return False
    return all(
        _re.search(pattern, key) is not None for key in state_dict if "adapter" in key
    )


def _find_matched_v0_adapter_key(
    v1_key: str, v1_state_dict: _Mapping[str, _Any]
) -> str:
    # pylint:disable=line-too-long,unused-argument
    """
    Utility function to map v1 adapter keys to v0 style.
    Example 1 (Single LoRA):
        v0: layers.layer_0.attention.qkv_transform.adapter.lora_Wq.a_transpose
        v1: layers.layer_0.attention.qkv_transform.adapters.my_lora.lora_0.a_transpose
    Example 2 (Multiple/Composite LoRA):
        v0: layers.layer_0.attention.qkv_transform.adapter.lora_Wq.a_transpose
        v1: layers.layer_0.attention.qkv_transform.adapters.<parent_adapter_id>.<child_adapter_id>.lora_0.a_transpose
    Example 3 (Multiple/Composite LoRA):
        v0: layers.layer_0.attention.qkv_transform._wrapped.adapter.lora_Wk.a_transpose
        v1: layers.layer_0.attention.qkv_transform.adapters.<parent_adapter_id>.<child_adapter_id>.lora_1.a_transpose
    """
    v0_key = v1_key[: v1_key.find(".adapters.")]
    match = _re.search(r".adapters\.(\w+)\.([0-9]+)", v1_key)
    if match:
        depth = int(match.groups()[1]) + 1
        v0_key += "._wrapped" * (depth - 1)
    v0_key += ".adapter"
    if ".lora_" in v1_key:
        # LoRAFusedMultiOutputLinear adapter presents
        v0_key += ".lora_Wq." if ".lora_0." in v1_key else ""
        v0_key += ".lora_Wk." if ".lora_1." in v1_key else ""
        v0_key += ".lora_Wv." if ".lora_2." in v1_key else ""
    else:
        v0_key += "."
    return v0_key + v1_key.split(".")[-1]


def _find_adapter_names(v1_key: str, v1_state_dict: _Mapping[str, _Any]) -> _List[str]:
    """
    Helper function which finds all adapters in the state_dict
    which are applied to the same layer as v1_key.
    """
    layer_prefix = (
        v1_key[: v1_key.find(".wrapped.")]
        if ".wrapped." in v1_key
        else v1_key[: v1_key.find(".adapters.")]
    )
    adapter_key_prefix = layer_prefix + ".adapters."
    adapter_names = set()
    for key in v1_state_dict:
        if adapter_key_prefix in key:
            modified_key = key[len(adapter_key_prefix) :]
            # check LoRALinear
            pattern = _re.compile(r"\.[ab]_transpose")
            match = _re.search(pattern, modified_key)
            if match:
                adapter_names.add(modified_key[: match.start()])
                continue
            # check LoRAFusedMultiOutputLinear
            pattern = _re.compile(r"\.lora_([0-2])\.[ab]_transpose")
            match = _re.search(pattern, modified_key)
            if match:
                adapter_names.add(modified_key[: match.start()])
                continue
    return sorted(list(adapter_names))


def _find_matched_v0_wrapped_key(
    v1_key: str, v1_state_dict: _Mapping[str, _Any]
) -> str:
    """
    Utility function to map v1 wrapped layer keys to v0 style.
    Example 1 (Single LoRA):
        v0: layers.layer_0.attention.qkv_transform._wrapped.fused_linear.weight
        v1: layers.layer_0.attention.qkv_transform.wrapped.fused_linear.weight
    Example 2 (Multiple/Composite LoRA):
        v0: layers.layer_0.attention.qkv_transform._wrapped._wrapped.fused_linear.weight
        v1: layers.layer_0.attention.qkv_transform.wrapped.fused_linear.weight
    """
    adapter_names = _find_adapter_names(v1_key, v1_state_dict)
    adapter_names = set(_re.sub(r"lora_([0-9]+)", "X", name) for name in adapter_names)
    num_adapters = max(1, len(adapter_names))
    return (
        v1_key[: v1_key.find(".wrapped.")]
        + "._wrapped" * num_adapters
        + v1_key[v1_key.find(".wrapped.") + len(".wrapped") :]
    )


def _find_matched_v0_key(v1_key: str, v1_state_dict: _Mapping[str, _Any]) -> str:
    if ".adapters." in v1_key:
        return _find_matched_v0_adapter_key(v1_key, v1_state_dict)
    if ".wrapped." in v1_key:
        return _find_matched_v0_wrapped_key(v1_key, v1_state_dict)
    return v1_key


def _find_matched_v1_key(
    v0_key: str, v1_adapter_id: str, use_parallel_stack_composition: bool
) -> str:
    if ".adapter." in v0_key:
        v1_key = v0_key[: v0_key.find(".adapter.")]
        if use_parallel_stack_composition:
            v1_adapter_idx = v0_key.count("._wrapped")
            v1_key = v1_key.replace("._wrapped", "")
            v1_key += f".adapters.{v1_adapter_id}.{v1_adapter_idx}"
        else:
            v1_key += f".adapters.{v1_adapter_id}"
        if ".lora_W" in v0_key:
            # LoRAFusedMultiOutputLinear adapter presents
            v1_key += ".lora_0." if ".lora_Wq." in v0_key else ""
            v1_key += ".lora_1." if ".lora_Wk." in v0_key else ""
            v1_key += ".lora_2." if ".lora_Wv." in v0_key else ""
        else:
            v1_key += "."
        return v1_key + v0_key.split(".")[-1]
    if "._wrapped." in v0_key:
        return v0_key.replace("._wrapped", "", v0_key.count("._wrapped") - 1).replace(
            "._wrapped", ".wrapped"
        )
    return v0_key


def create_v1_state_dict(
    v0_state_dict: _Mapping[str, _Any], v1_adapter_id: str
) -> _Mapping[str, _Any]:
    """
    Create a v1 adapted model state_dict from a v0 ``tamm`` model.
    """
    _warnings.warn(
        "This conversion API assumes the v1 adapted model "
        "which loads the converted state_dict will only have one adapter "
        f"with the input adapter_id: {v1_adapter_id}."
    )
    v1_state_dict = _copy(v0_state_dict)
    use_parallel_stack_composition = any(
        "._wrapped.adapter" in v0_key for v0_key in v0_state_dict
    )
    for v0_key in v0_state_dict:
        v1_key = _find_matched_v1_key(
            v0_key, v1_adapter_id, use_parallel_stack_composition
        )
        if v1_key in v1_state_dict:
            continue
        v1_state_dict[v1_key] = v1_state_dict.pop(v0_key)
    return v1_state_dict


def maybe_rekey_v0_key(
    v1_state_dict: _Mapping[str, _Any], v0_state_dict: _Mapping[str, _Any]
) -> _Mapping[str, _Any]:
    """
    Load a v0 adapted ``tamm`` model state_dict.
    Only attention and feedforward layers
    adapted with LoRA adapters can be converted.
    """
    remaining_v0_state_dict_keys = set(v0_state_dict.keys())
    new_v1_state_dict = _OrderedDict()

    for v1_key in v1_state_dict:
        v0_key = _find_matched_v0_key(v1_key, v1_state_dict)
        if v0_key in v0_state_dict:
            new_v1_state_dict[v1_key] = v0_state_dict[v0_key]
            remaining_v0_state_dict_keys.remove(v0_key)
    # add the extra keys in v0 which v1 doesn't expect to v1 state dict so it can be handled
    # by strict Flag
    if len(remaining_v0_state_dict_keys) > 0:
        for v0_key in remaining_v0_state_dict_keys:
            new_v1_state_dict[v0_key] = v0_state_dict[v0_key]
    return new_v1_state_dict


def maybe_rekey_state_dict(
    model: _nn.Module, state_dict: _Mapping[str, _Any]
) -> _Mapping[str, _Any]:
    """
    Helper function to align state dict keys with a module's param names
    when one of the module or state dict has adapters but the other does not.
    Returns a copy of the state dict where keys have been updated to
    include or exclude ".wrapped" if this
    aligns the key with a module param name.
    """
    state_dict = _copy(state_dict)
    module_keys = [
        key.replace(_TORCH_CHECKPOINT_ACTIVATION_KEY, "") for key in model.state_dict()
    ]

    for module_name, module in model.named_modules():
        if isinstance(module, _AdaptedLayer):
            for key in module.state_dict().keys():
                stripped_module_name = module_name.replace(
                    _TORCH_CHECKPOINT_ACTIVATION_KEY, ""
                )
                stripped_key = key.replace(_TORCH_CHECKPOINT_ACTIVATION_KEY, "")
                original_key = stripped_module_name + "." + stripped_key
                unwrapped_key = stripped_module_name + stripped_key[len("wrapped") :]
                if (
                    unwrapped_key in state_dict
                    and original_key not in state_dict
                    and unwrapped_key not in module_keys
                ):
                    state_dict[original_key] = state_dict.pop(unwrapped_key)

    def create_unwrapped_keys_helper(splitted_keys, prev_unwrapped_keys) -> _List[str]:
        """
        Helper function to generate unwrapped keys considering
        the possibility of any nested AdaptedLayer existing.
        """
        if not splitted_keys:
            return prev_unwrapped_keys
        cur_unwrapped_keys = []
        for prev_unwrapped_key in prev_unwrapped_keys:
            cur_unwrapped_keys.append(prev_unwrapped_key + splitted_keys[0])
            cur_unwrapped_keys.append(
                prev_unwrapped_key + ".wrapped" + splitted_keys[0]
            )
        return create_unwrapped_keys_helper(splitted_keys[1:], cur_unwrapped_keys)

    for key in list(state_dict.keys()):
        unwrapped_keys = []
        splitted_keys = key.split(".wrapped")
        if len(splitted_keys) == 1:
            unwrapped_keys = [key]
        elif len(splitted_keys) == 2:
            unwrapped_keys = ["".join(splitted_keys)]
        else:
            unwrapped_keys.extend(
                create_unwrapped_keys_helper(splitted_keys[1:], [splitted_keys[0]])
            )

        for unwrapped_key in unwrapped_keys:
            if (
                unwrapped_key in module_keys
                and unwrapped_key not in state_dict
                and key not in module_keys
            ):
                state_dict[unwrapped_key] = state_dict.pop(key)
                break

    return state_dict


def remap_adapter_state_dict_keys(
    model: _nn.Module, state_dict: _Mapping[str, _Any]
) -> _Mapping[str, _Any]:
    if _rc.adapters_implementation == "v1":
        if is_v0_lora_state_dict(state_dict):
            # Convert v0 style state_dict to v1.
            state_dict = maybe_rekey_v0_key(
                v1_state_dict=model.state_dict(), v0_state_dict=state_dict
            )
        # Handling possible extra/missing ".wrapped" in state_dict keys.
        state_dict = maybe_rekey_state_dict(model, state_dict)
    return state_dict


def get_linear_adapter(key: str) -> "_LayerAdapter":
    """
    Backward compatible replacement for :func:`adapters_v0.get_linear_adapter`.
    This method only takes input key as "lora" or "lora-linear".
    Where "lora" maps to `LoRAFusedMultiOutputLinear.Config` and "lora-linear" maps to `LoRA.Config`.
    """
    if key not in {"lora", "lora-linear"}:
        raise RuntimeError(
            "`_adapter_v1.get_linear_adapter` accepts either 'lora' or 'lora-linear'."
        )
    if key == "lora":
        return _LoRAFusedMultiOutputLinear.Config
    return _LoRA.Config


def split_adapted_weights(state_dict: dict) -> _Dict[str, _Dict]:
    """
    Utility function that splits adapter fine-tuned model weights into backbone and
    adapter weights.

    Args:
        state_dict: State dictionary of the complete model (backbone+adapter)

    Returns: dictionary with keys "backbone_state_dict", and "adapter_state_dict"

    """

    def _is_adapter_key(name):
        return "adapters" in name

    def _normalize(name):
        return name.replace(".wrapped", "")

    backbone = {k: v for k, v in state_dict.items() if not _is_adapter_key(k)}
    adapter = {k: v for k, v in state_dict.items() if _is_adapter_key(k)}

    backbone = {_normalize(k): v for k, v in backbone.items()}
    return {
        "backbone_state_dict": backbone,
        "adapter_state_dict": adapter,
    }


def merge_adapters(module, remove=True):
    """
    Merge active adapters into the backbone modules.
    If there's no active adapters in the passed in module,
    this function will become a no-op.

    Args:
        module (:obj:`_nn.Module`): the model with lora adapters.
        remove (:obj:`bool`): Flag to remove the extra adapter post-merge.
    """
    # pylint: disable=import-outside-toplevel
    from tamm._adapters_v1.adapter_api import (
        get_all_active_adapter_ids,
        is_adapter_initialized,
        merge_adapter,
        uninit,
    )

    active_adapter_ids = get_all_active_adapter_ids(module)

    if not is_adapter_initialized(module) or len(active_adapter_ids) == 0:
        _warnings.warn(
            "No active adapters found. Nothing to be merged. This `merge_adapters` will be a NO-OP."
        )

    for adapter_id in active_adapter_ids:
        merge_adapter(module, adapter_id)
    if remove:
        module = uninit(module, force=True)

    return module


def get_num_adapter_params(model: _nn.Module, only_trainable: bool = False):
    """
    Utility function to get the number of adapter parameters in the given model.

    Args:
        model (:obj:`_nn.Module`): The input adapted model.
        only_trainable (:obj:`bool`):
            Flag to indicate if only trainable parameters are counted.
    """
    if only_trainable:
        return sum(
            _torch_utils.get_num_trainable_params(module.adapters)
            for module in model.modules()
            if isinstance(module, _AdaptedLayer)
        )
    return sum(
        _torch_utils.get_num_params(module.adapters)
        for module in model.modules()
        if isinstance(module, _AdaptedLayer)
    )


def find_adapter_ids_from_state_dict(state_dict: _StateDictType) -> _List[str]:
    """
    Finds all possible adapter_ids in the state dict, by finding all keys
    which match *.adapters.<adapter_id>.* pattern.
    """
    adapter_ids = set()
    for key in state_dict:
        pattern = _re.compile(r"(?<=\.adapters\.)[^.]+(?=\.|$)")
        match = _re.search(pattern, key)
        if match:
            adapter_ids.add(match.group())
            continue
    return sorted(list(adapter_ids))



import logging as _logging

from tamm import _helpers
from tamm.utils import partial as _partial
from tamm.utils.json import JSONSerializableMixin as _JSONSerializableMixin

_logger = _logging.getLogger(__name__)


class _BaseConfig(
    _partial.DataclassedPartial,
    _JSONSerializableMixin,
):
    """
    The base class for :class:`.ModelConfig` and :class:`.LayerConfig`.
    Not to be instantiated directly.
    """

    @property
    def has_adapters(self):
        sub_modules = ["decoder", "image_tokenizer"]
        for sub_module in sub_modules:
            if hasattr(self, sub_module) and getattr(self, sub_module).has_adapters:
                return True
        if not hasattr(self, "adapters"):
            return False
        # pylint: disable=no-member
        return isinstance(self.adapters, dict) and len(self.adapters) > 0

    def _create(self, *override_args, **override_kwargs):
        """
        Creates and returns a configured instance of the module.

        Args:
            override_args: Optional positional arguments to override args specified in
                the config.  These args replace the first ``len(override_args)``
                positional args (and *all* varargs if ``override_args`` contains
                varargs) in the layer's :meth:`~.ModelMixin.create_builder` method.
            override_kwargs: Optional keyword override arguments.  These arguments
                replace any additional named arguments not overridden by
                ``override_args``.

        Returns:
            The newly created layer.
        """
        builder = self.create_builder(*override_args, **override_kwargs)
        return builder.build()

    def create_builder(self, *override_args, **override_kwargs):
        """
        Creates and returns a configured module builder. This method accepts the same
        arguments as ``create_layer()`` or ``create_model()``.
        """
        return self(*override_args, **override_kwargs)

    @classmethod
    def _create_subclass_from_module_cls(cls, layer_cls):
        """
        Creates a :class:`.LayerConfig` for a layer class which implements
        create_builder(*args, **kwargs)
        """
        layer_name = layer_cls.__name__
        config_name = f"{layer_name}Config"
        config_cls = cls.create_subclass(
            target_callable=layer_cls.create_builder,
            name=config_name,
            module_path=layer_cls.__module__,
        )
        config_cls.__doc__ = (
            f"A :class:`.LayerConfig` subclass for configuring :class:`.{layer_name}` "
            f"layers.  Use the alias :attr:`.{layer_name}.Config` to access this "
            f"class.  Please check :meth:`.{layer_name}.create_builder` for more "
            f"details about the  signature."
        )
        return config_cls

    def _to_json_dict_impl(self):
        result = _helpers.dataclass_to_dict(
            self.configured_args,
            omit_defaults=True,
            # only save non-default values for forward compatibility (if we include
            # default values, older tamm versions cannot load configs from newer
            # tamm versions that introduce new options)
        )
        result.pop("pretrained", None)  # specify pretrained=True in code, not config
        return result

    @classmethod
    def _from_json_dict_impl(cls, **obj):
        obj.pop("__versioning_info__", None)
        is_pretrained = obj.pop("pretrained", False)
        if is_pretrained:
            _logger.warning(
                "Dropping pretrained=True option when loading LayerConfig from JSON. "
                "Please specify this argument explicitly after loading the config."
            )
        return super()._from_json_dict_impl(**obj)


class LayerConfig(
    _BaseConfig,
    json_namespace="layers",
):
    """
    The base class for layer configs.  A :class:`.LayerConfig` is a
    :class:`.DataclassedPartial` with the layer class as the target callable.

    For usage examples, please see the layer
    configuration :ref:`user guide <configure_models>`.  Layer developers should use the
    ``.LayerMixin`` to create :class:`.LayerConfig` subclasses for new layers.
    """

    def create_layer(self, *override_args, **override_kwargs):
        """
        Creates and returns a configured instance of the layer.

        Args:
            override_args: Optional positional arguments to override args specified in
                the config.  These args replace the first ``len(override_args)``
                positional args (and *all* varargs if ``override_args`` contains
                varargs) in the layer's :meth:`~.ModelMixin.create_builder` method.
            override_kwargs: Optional keyword override arguments.  These arguments
                replace any additional named arguments not overridden by
                ``override_args``.

        Returns:
            The newly created layer.
        """
        return self._create(*override_args, **override_kwargs)



from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
from torch.nn import functional as _F

from tamm import _helpers
from tamm.utils import _torch_compatibility


def rms_norm(
    tensor: _torch.Tensor,
    normalized_shape: _Tuple[int],
    *,
    weight: _torch.Tensor = None,
    bias: _torch.Tensor = None,
    eps: float = 1e-05,
    cast_dtype: _Optional[_torch.dtype] = _torch.float32,
) -> _torch.Tensor:
    """Applies RMS normalization."""
    input_dtype = tensor.dtype
    with _helpers.autocast_disabled(tensor.device):
        if cast_dtype is not None:
            tensor = tensor.type(cast_dtype)
        tensor = _torch_compatibility.rms_norm(
            tensor, normalized_shape=normalized_shape, weight=weight, eps=eps
        )
        if bias is not None:
            tensor = tensor + bias
        return tensor.type(input_dtype)


def batched_rms_norms(
    tensor: _torch.Tensor,
    normalized_shape: _Tuple[int],
    *,
    weight: _torch.Tensor = None,
    bias: _torch.Tensor = None,
    eps: float = 1e-05,
    cast_dtype: _Optional[_torch.dtype] = _torch.float32,
) -> _torch.Tensor:
    """
    Applies RMS normalization.  This the same a :func:`rms_norm`, but the weight
    and bias tensors can have more dimensions than ``len(normalized_shape)``.
    """
    input_dtype = tensor.dtype
    with _helpers.autocast_disabled(tensor.device):
        if cast_dtype is not None:
            tensor = tensor.type(cast_dtype)
        tensor = _torch_compatibility.rms_norm(
            tensor, normalized_shape=normalized_shape, eps=eps
        )
        if weight is not None and bias is not None:
            tensor = _torch.addcmul(bias, tensor, weight)
        elif weight is not None:
            tensor = tensor * weight
        elif bias is not None:
            tensor = tensor + bias
        return tensor.type(input_dtype)


def layer_norm(
    tensor: _torch.Tensor,
    normalized_shape: _List[int],
    *,
    weight: _Union[_torch.Tensor, None] = None,
    bias: _Union[_torch.Tensor, None] = None,
    eps: float = 1e-05,
    cast_dtype: _Optional[_torch.dtype] = _torch.float32,
) -> _torch.Tensor:
    with _helpers.autocast_disabled(tensor.device):
        input_dtype = tensor.dtype
        if cast_dtype is not None:
            tensor = tensor.type(cast_dtype)
        if weight is not None:
            weight = weight.type_as(tensor)
        if bias is not None:
            bias = bias.type_as(tensor)
        result = _torch.nn.functional.layer_norm(
            tensor, normalized_shape=normalized_shape, weight=weight, bias=bias, eps=eps
        )
        return result.type(input_dtype)


def batch_norm(
    tensor: _torch.Tensor,
    *,
    running_mean: _Union[_torch.Tensor, None],
    running_var: _Union[_torch.Tensor, None],
    weight: _Union[_torch.Tensor, None] = None,
    bias: _Union[_torch.Tensor, None] = None,
    training: bool = False,
    momentum: float = 0.1,
    eps: float = 1e-05,
    cast_dtype: _Optional[_torch.dtype] = _torch.float32,
) -> _torch.Tensor:
    with _helpers.autocast_disabled(tensor.device):
        input_dtype = tensor.dtype
        if cast_dtype is not None:
            tensor = tensor.type(cast_dtype)
        result = _torch.nn.functional.batch_norm(
            tensor,
            running_mean=running_mean,
            running_var=running_var,
            weight=weight,
            bias=bias,
            training=training,
            momentum=momentum,
            eps=eps,
        )
        return result.type(input_dtype)


# pylint: disable-next=redefined-builtin
def swi_glu(gate_input, input):
    """
    GLU variant that uses Swish nonlinear function in place of sigmoid.

    Args:
        gate_input (:obj:`torch.Tensor`): The input to the silu function.
        input (:obj:`torch.Tensor`): The input to scale by the silu output.

    Returns:
        The activation tensor.
    """
    return input * _F.silu(gate_input)


def l2_norm(
    tensor: _torch.Tensor,
    eps: float = 1e-05,
    cast_dtype: _Optional[_torch.dtype] = _torch.float32,
) -> _torch.Tensor:
    if cast_dtype is not None:
        cast_tensor = tensor.type(cast_dtype)
    else:
        cast_tensor = tensor

    return _F.normalize(cast_tensor, dim=-1, eps=eps)


def rearrange_embeddings_to_channels_first(embeddings, *, unflattened_shape):
    embeddings = embeddings.movedim(1, -1)
    return embeddings.reshape(*embeddings.shape[:-1], *unflattened_shape)


def rearrange_embeddings_from_channels_first(tensor):
    return tensor.flatten(start_dim=2).movedim(1, -1)


def geglu(gate_input, input):  # pylint: disable=redefined-builtin
    """
    GLU variant that uses GELU function in place of sigmoid.

    Args:
        gate_input (:obj:`torch.Tensor`): The input to the GELU function.
        input (:obj:`torch.Tensor`): The input to scale by the GELU output.

    Returns:
        The activation tensor.
    """
    # pylint: disable=not-callable
    return input * _F.gelu(gate_input, approximate="tanh")


def add_batch_dim(tensor: _torch.Tensor, batch_size: int = 1) -> _torch.Tensor:
    """
    Adds a new first dimension to ``tensor`` and then repeats it ``batch_size`` times
    along that dimension.
    """
    tile_shape = (batch_size,) + (1,) * tensor.ndim
    return tensor.tile(tile_shape)


def maybe_flatten_sequence(
    sequence: _Union[_torch.Tensor, _Dict[str, _torch.Tensor]], end_dim: int = -1
) -> _torch.Tensor:
    """
    If ``sequence`` is a tensor, this function flattens the dimensions between ``1``
    and ``end_dim``.  If ``sequence`` is a :obj:`dict`, this function flattens each
    of its values and then concatenates the flat tensors along dimension 1 in the order
    that they appear in the :obj:`dict`.  If ``sequence`` is neither a tensor nor a
    :obj:`dict`, this function simply returns the ``sequence`` unchanged.
    """
    if _torch.is_tensor(sequence):
        end_dim = end_dim % sequence.ndim
        if end_dim > 1:
            return sequence.flatten(start_dim=1, end_dim=end_dim)
        return sequence

    if not isinstance(sequence, dict):
        return sequence

    sequence = [
        maybe_flatten_sequence(tensor, end_dim=end_dim) for tensor in sequence.values()
    ]
    return _torch.cat(sequence, dim=1)


def maybe_unflatten_sequence(sequence, *, original, end_dim: int = -1):
    """
    This function performs the inverse of :func:`maybe_flatten_sequence`.  It unflattens
    ``sequence`` to match the structure of ``original``.  Here ``original`` may be
    a :obj:`dict` or tensor.
    """
    if _torch.is_tensor(original):
        spatial_shape = original.shape[1 : (end_dim % original.ndim) + 1]
        if len(spatial_shape) == 1:
            return sequence
        return sequence.reshape(sequence.shape[0], *spatial_shape, *sequence.shape[2:])

    if not isinstance(original, dict):
        return sequence

    original_tensors = list(original.values())
    flat_original_tensors = [
        maybe_flatten_sequence(tensor, end_dim=end_dim) for tensor in original.values()
    ]
    seq_lens = [o_i.size(1) for o_i in flat_original_tensors]
    sequence = _torch.split(sequence, seq_lens, dim=1)
    sequence = [
        maybe_unflatten_sequence(e_i, original=o_i, end_dim=end_dim)
        for e_i, o_i in zip(sequence, original_tensors)
    ]
    return dict(zip(original, sequence))


def maybe_flatten_embeddings(
    embeddings: _Union[_torch.Tensor, _Dict[str, _torch.Tensor]]
) -> _torch.Tensor:
    """
    A convenience function for calling :func:`maybe_flatten_sequence` with
    ``end_dim=-2``.  This is helpful for flattening spatial embeddings with shape
    ``(batch_size, height, width, dim)`` into shape ``(batch_size, sequence_len, dim)``.
    """
    return maybe_flatten_sequence(embeddings, end_dim=-2)


def maybe_unflatten_embeddings(
    embeddings: _torch.Tensor,
    *,
    original: _Union[_torch.Tensor, _Dict[str, _torch.Tensor]],
) -> _Union[_torch.Tensor, _Dict[str, _torch.Tensor]]:
    """
    A convenience function for calling :func:`maybe_unflatten_sequence` with
    ``end_dim=-2``.
    """
    return maybe_unflatten_sequence(embeddings, original=original, end_dim=-2)


def inverse_permute(tensor: _torch.Tensor, perm: _List[int]):
    """
    A function that performs the inverse of :func:`torch.permute`.

    Args:
        tensor (:obj:`torch.Tensor`): Tensor to unpermute.
        perm (:obj:`List[int]`): Int list specifying the original axis permutation.

    Returns:
        The inverse permutation of tensor.
    """
    inv_perm = [perm.index(idx) for idx in range(len(perm))]
    assert len(inv_perm) == len(perm)
    return _torch.permute(tensor, inv_perm)


def stack(*tensors, dim=0):
    """
    This function operates similarly to torch.stack, but accepts multiple tensors as
    input arguments instead of a list of tensors. It can be frequently used to fuse
    parameters because the 'to_tamm' and 'from_tamm' functions in ParamMapper
    do not accept lists as input.
    """
    return _torch.stack(tensors, dim=dim)


def relaxed_one_hot(
    tensor: _torch.Tensor, *, num_classes: int, dtype: _torch.dtype = _torch.bool
):
    """
    This function is similar to :func:`torch.nn.functional` except for the following
    differences:

    1. The function ignores indices outside the range ``[0, num_classes)`` rather
       than raising an error.
    2. The dtype of the result defaults to ``bool``, and it is also configurable.
    3. The ``num_classes`` argument is required.

    Args:
        tensor (:obj:`torch.Tensor`): An integer tensor of indices.
        num_classes (:obj:`int`): The embedding size of the resulting one-hot encodings.
        dtype (:obj:`torch.dtype`): The result dtype.

    Returns:
        A binary-valued tensor with shape ``(*tensor.shape, num_classes)``.  Each
        entry takes value ``1`` if its index corresponds to an index in ``tensor``
        (and ``0`` otherwise).
    """
    input_shape = tensor.shape
    if tensor.device.type == "mps":
        tensor = tensor.flatten()  # due to an mps bug as of torch 2.5

    arange = _torch.arange(num_classes, device=tensor.device, dtype=tensor.dtype)
    result = tensor[..., None] == arange

    if tensor.device.type == "mps":
        result = result.reshape(*input_shape, -1)

    return result.type(dtype)


def cumsum(tensor: _torch.Tensor, dim: int) -> _torch.Tensor:
    """
    This is the same as :func:`torch.cumsum` but it is much faster on CUDA
    for important use cases.
    """
    should_move_dim = dim not in (-1, tensor.ndim - 1)
    if should_move_dim:
        tensor = tensor.movedim(dim, -1)
    result = tensor.cumsum(dim=-1)
    if should_move_dim:
        result = result.movedim(-1, dim)
    return result


def segment_matmul(segmented_input, group_sizes, weight):
    """
    Segment matrix multiplication:
    Perform a segmented or jagged/nested matrix multiplication against input,
    with different lengths of batched inputs against correspondingly indexed weights.
    Written for the purpose of dropless MoE mode.
    Example: (t=tracks/vectorized dim, b=batch, d_in=hidden in dim, d_out=hidden out dim)
    - `segmented_input` with shape (t, b, d_in)
    - `group_sizes` with shape (t, e)
    - `weight` with shape (t, e, d_in, d_out)
    - Returns:
          tensor of shape (t, b, d_out), where
          cum_group_sizes = cumsum(group_sizes, dim=1)
          result = stack([cat([segmented_input[i][0:cum_group_sizes[i][0]] @ weight[i,0]
                               segmented_input[i][cum_group_sizes[i][0]:cum_group_sizes[i][1]] @ weight[i,1]
                               ...
                              ]
                              for i in tracks
                          ])
    Args:
        segmented_input: The input tensor
        group_sizes: sizes of each group of `segmented_input` to matmul against `weight`
        weight: Weight tensor
    Returns:
        Output of segmented matrix multiplication
    """

    # Ensure that we're executing "vectorized segment matmul" (i.e: a PTT MoE activation layer)
    # A normal non-vectorized segment matmul can be performed with an explicit tracks=1.
    # input expected to be of shape (tracks, batch*seq*top_k, hidden_dim) in MoE case
    if len(segmented_input.shape) != 3:
        raise ValueError(
            "segment_matmul currently only accepts 3-dimensional inputs (parallel-track op)"
        )

    # For each track, perform segment matmul
    tracks = segmented_input.shape[0]
    group_sizes_list = group_sizes.tolist()
    track_results = []
    for track in range(tracks):
        # Following contains a list of tensors to pass to each expert, of length experts
        track_experts_tensors = _torch.split(
            segmented_input[track], group_sizes_list[track]
        )

        # Perform feedforward per expert
        for idx, expert_tensor in enumerate(track_experts_tensors):
            track_results.append(_torch.matmul(expert_tensor, weight[track, idx]))

    return _torch.cat(track_results).reshape((tracks, -1, weight.shape[-1]))


def crop(tensor: _torch.Tensor, *, shape: _Tuple[int, ...]) -> _torch.Tensor:
    """
    Truncates tensors with shape ``(batch_size, num_channels, *spatial_shape)``
    at top left corner to tensors with shape ``(batch_size, num_channels, *shape)``.
    """
    tensor_spatial_shape = tensor.shape[2:]
    if tensor_spatial_shape == shape:
        return tensor
    if any(i_dim > j_dim for i_dim, j_dim in zip(shape, tensor_spatial_shape)):
        raise ValueError(
            "Each dimension of truncation_shape must not be larger than the "
            "corresponding dimension in the tensor to be truncated."
        )
    slices = (slice(None), slice(None)) + tuple(slice(dim) for dim in shape)
    tensor = tensor[slices]
    return tensor


def soft_cap(tensor: _torch.Tensor, *, cap: float) -> _torch.Tensor:
    """
    Computes ``cap * tanh(tensor / cap)``, which is a smooth and differentiable way
    to cap the values of a tensor.

    Raises: ValueError: If ``cap`` is non-positive.
    """

    if cap <= 0:
        raise ValueError(f"soft_cap requires cap > 0, but cap is {cap}")
    return cap * _torch.tanh(tensor / cap)



import warnings

import wrapt

warnings.filterwarnings("always", category=DeprecationWarning, module=r"^tamm\..*")


def deprecate(alternative=None, *, name=None):
    @wrapt.decorator
    def wrapper(wrapped, _instance, args, kwargs):
        nonlocal name

        if alternative is not None:
            alternative_message = f" Please use {alternative} instead."
        else:
            alternative_message = ""
        if name is None:
            name = wrapped.__name__

        msg = (
            f"{name} will be removed in a future version of tamm."
            f"{alternative_message}"
        )

        warnings.warn(msg, category=DeprecationWarning, stacklevel=2)

        return wrapped(*args, **kwargs)

    return wrapper


def deprecation(message: str):
    warnings.warn(message, category=DeprecationWarning, stacklevel=2)



from tamm import preprocessors
from tamm._cli._listers._base import BaseLister


class PreprocessorLister(BaseLister):
    name = "preprocessors"

    def print(self):
        print("\n".join(preprocessors.list()))



import importlib.metadata as _importlib_metadata
import logging as _logging
from typing import Union as _Union

import packaging.requirements as _packaging_requirements
from packaging.version import parse as _parse_version

from tamm._version import __version__ as _TAMM_VERSION
from tamm.runtime_configuration import rc as _rc

_logger = _logging.getLogger(__name__)


def _get_tamm_dist() -> _importlib_metadata.Distribution:
    """Returns the tamm distribution"""
    tamm_ver = _parse_version(_TAMM_VERSION)
    distributions = [
        dist
        for dist in _importlib_metadata.distributions()
        if _parse_version(dist.version) == tamm_ver
    ]

    for dist in distributions:
        try:
            top_level_text = dist.read_text("top_level.txt")
        except Exception as e:
            _logger.debug(f"Caught exception reading top_level.txt for {dist}: {e}")
            continue

        if _rc.PROJECT_SLUG in top_level_text.split():
            return dist

    raise RuntimeError("Unable to locate the tamm distribution")


def _get_specifier_set(
    dist: _importlib_metadata.Distribution, *, package_name
) -> _Union[_packaging_requirements.SpecifierSet, None]:
    """
    Returns the specfier set for the ``dist`` requirement with name ``package_name``.
    Returns ``None`` if the package is not a requirement
    """
    requirements = (_packaging_requirements.Requirement(req) for req in dist.requires)
    for requirement in requirements:
        if requirement.name == package_name:
            return requirement.specifier
    return None


def _get_min_version(
    specifier_set: _packaging_requirements.SpecifierSet,
) -> _Union[str, None]:
    """
    Returns the version lower bound for a requirement and ``None`` if it does not exist.
    """
    for specifier in specifier_set:
        if specifier.operator == ">=":
            return specifier.version
    return None


def _get_python_requirement() -> _packaging_requirements.SpecifierSet:
    """Returns the version specfier set for tamm's Python requirement."""
    dist = _get_tamm_dist()
    python_requirement = dist.metadata["Requires-Python"]
    return _packaging_requirements.SpecifierSet(python_requirement)


def get_min_python_version() -> str:
    """Returns the Python version lower-bound for the tamm package."""
    spec_set = _get_python_requirement()
    return _get_min_version(spec_set)


def _get_torch_requirement() -> _packaging_requirements.SpecifierSet:
    """Returns the version specfier set for tamm's torch requirement."""
    dist = _get_tamm_dist()
    return _get_specifier_set(dist, package_name="torch")


def get_min_torch_version() -> str:
    """Returns the torch version lower-bound for the tamm package."""
    spec_set = _get_torch_requirement()
    return _get_min_version(spec_set)



import collections as _collections

import torch as _torch
from torch import nn as _nn


def _compile_friendly_get_parameter(module, parameter_name):
    """similar to module.get_parameter(parameter_name) but works with torch.compile"""
    module_or_param = module
    for name in parameter_name.split("."):
        module_or_param = getattr(module_or_param, name)
    return module_or_param


class TieableParamsDict:
    """
    A replacement of :class:`OrderedDict` for holding parameters in the ``_parameters``
    attribute of a :obj:`torch.nn.Module`.  This class implements special behavior for
    sharing parameters between modules to ensure that the parameters do not become
    untied.  It is also compatible with :func:`torch.compile` and FSDP.
    """

    def __init__(self, *args, **kwargs):
        self._params = _collections.OrderedDict(*args, **kwargs)
        self._tied_parameters = {}

    def register_tied_parameter(
        self, *, name: str, tied_module: _nn.Module, tied_param_name: str
    ):
        """
        Register a parameter from another module.

        Args:
            name (:obj:`str`): The name of the parameter to register.
            tied_module (:obj:`nn.Module`): The other module that owns the tied
                parameter to register.
            tied_param_name (:obj:`str`): The name of the parameter as it is
                registered on ``tied_module``.
        """
        self._tied_parameters[name] = (tied_module, tied_param_name)
        self._update_tied_params()

    def get_tied_value(self, key: str) -> _torch.Tensor:
        """
        A method for accessing the tied parameters.

        Args:
            key (:obj:`str`): The tied parameter name.

        Returns:
            The tied parameter.
        """
        if key in self._tied_parameters:
            module, name = self._tied_parameters[key]
            return _compile_friendly_get_parameter(module, name)
        return self._params[key]

    def _update_tied_params(self):
        for name, (tied_module, tied_param_name) in self._tied_parameters.items():
            value = _compile_friendly_get_parameter(tied_module, tied_param_name)
            if isinstance(value, _nn.Parameter):
                self._params[name] = value
            else:
                # As of torch 2.5, FSDP at times replaces each nn.Parameter with a
                # vanilla torch.Tensor, in which case, the tensor should not show
                # up in the parameters dict.
                self._params.pop(name, None)

    def __eq__(self, other):
        self._update_tied_params()
        return self._params.__eq__(other)

    def __ne__(self, other):
        self._update_tied_params()
        return self._params.__ne__(other)

    def __iter__(self):
        self._update_tied_params()
        return self._params.__iter__()

    def __reversed__(self):
        self._update_tied_params()
        return self._params.__reversed__()

    def __getitem__(self, key):
        if key in self._tied_parameters:
            self._update_tied_params()
        return self._params.__getitem__(key)

    def __setitem__(self, key, value):
        if key in self._tied_parameters:
            return
        self._params.__setitem__(key, value)

    def __delitem__(self, key):
        if key in self._tied_parameters:
            return
        self._params.__delitem__(key)

    @classmethod
    def fromkeys(cls, iterable, value=None):
        params = {key: value for key in iterable}
        return cls(params)

    def setdefault(self, /, key, default=None):
        if key in self._tied_parameters:
            self._update_tied_params()
            return self[key]
        return self._params.setdefault(key, default=default)

    def pop(self, *args):
        key = args[0]
        if key in self._tied_parameters:
            self._update_tied_params()
            return self[key]
        return self._params.pop(*args)

    def popitem(self, last=True):
        self._update_tied_params()
        key, value = self._params.popitem(last=last)
        if key in self._tied_parameters:
            self._update_tied_params()
        return key, value

    def keys(self):
        self._update_tied_params()
        return self._params.keys()

    def values(self):
        self._update_tied_params()
        return self._params.values()

    def items(self):
        self._update_tied_params()
        return self._params.items()

    def clear(self):
        self._params.clear()
        self._update_tied_params()

    def move_to_end(self, key, last=True):
        self._update_tied_params()
        self._params.move_to_end(key, last=last)



from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple

import torch as _torch

from tamm.layers.transformer.kv_cache import common as _common


class VanillaKVCache(_common.BaseKVCache):
    """
    A vanilla implementation of :class:`.BaseKVCache`, which handles basic decoding
    scenarios.
    """

    def __init__(
        self,
        num_layers: int,
        batch_size: int,
        length: int,
        hidden_dim: int,
        *,
        device=None,
        dtype=None,
    ):
        self.keys_and_values = None
        self.positions = None
        self.token_types = None
        self.segment_ids = None
        self._init_tensors(
            num_layers=num_layers,
            batch_size=batch_size,
            length=length,
            hidden_dim=hidden_dim,
            device=device,
            dtype=dtype,
        )

    def _init_tensors(self, num_layers, batch_size, length, hidden_dim, device, dtype):
        self.keys_and_values = _torch.empty(
            num_layers, 2, length, batch_size, hidden_dim, device=device, dtype=dtype
        )

        self.segment_ids = _torch.ones(
            length, batch_size, device=device, dtype=_torch.int64
        )

        positions = _torch.arange(length, device=device, dtype=_torch.int64)
        self.positions = positions.tile((batch_size, 1)).T

        self.token_types = _torch.ones_like(self.segment_ids)

    @property
    def dtype(self) -> _torch.dtype:
        """The :obj:`torch.dtype` for keys and values stored in the cache."""
        return self.keys_and_values.dtype

    @property
    def device(self) -> _torch.device:
        """The device for keys and values in the cache."""
        return self.keys_and_values.device

    @property
    def num_layers(self) -> int:
        return self.keys_and_values.size(0)

    @property
    def batch_size(self) -> int:
        """The batch size of the cache contents."""
        return self.keys_and_values.size(3)

    @property
    def length(self) -> int:
        return self.keys_and_values.size(2)

    @property
    def hidden_dim(self) -> int:
        """The dimension of the keys and values in the cache."""
        return self.keys_and_values.size(4)

    def up_to_index(self, index: int) -> "VanillaKVCacheView":
        return VanillaKVCacheView(self, length=index)

    def resize(self, new_length: int):
        num_old_to_copy = min(self.length, new_length)
        old_keys_and_values = self.keys_and_values[:, :, :num_old_to_copy]
        old_segment_ids = self.segment_ids[:num_old_to_copy]
        old_positions = self.positions[:num_old_to_copy]
        old_token_types = self.token_types[:num_old_to_copy]

        self._init_tensors(
            num_layers=self.num_layers,
            batch_size=self.batch_size,
            length=new_length,
            hidden_dim=self.hidden_dim,
            device=self.device,
            dtype=self.dtype,
        )

        self.keys_and_values[:, :, :num_old_to_copy] = old_keys_and_values
        self.segment_ids[:num_old_to_copy] = old_segment_ids
        self.positions[:num_old_to_copy] = old_positions
        self.token_types[:num_old_to_copy] = old_token_types

    def to(self, *args, **kwargs) -> "VanillaKVCache":
        self.keys_and_values = self.keys_and_values.to(*args, **kwargs)
        self.segment_ids = self.segment_ids.to(self.keys_and_values.device)
        self.positions = self.positions.to(self.keys_and_values.device)
        self.token_types = self.token_types.to(self.keys_and_values.device)
        return self


class VanillaKVCacheView(_common.BaseKVCacheView):
    """
    A vanilla implementation of :class:`._common.BaseKVCacheView`, which handles basic
    decoding scenarios.

    Args:
        cache (:obj:`.KVCache`): A KV cache.
        length (:obj:`int`): The sequence length of the cache view.
    """

    def __init__(self, cache: VanillaKVCache, *, length: int):
        self.cache = cache
        self._length = length

    @property
    def length(self) -> int:
        return self._length

    @property
    def num_layers(self):
        return self.cache.num_layers

    @property
    def batch_size(self):
        """The batch size of the cache."""
        return self.cache.batch_size

    @property
    def dtype(self) -> _torch.dtype:
        return self.cache.dtype

    @property
    def device(self) -> _torch.device:
        return self.cache.device

    def write_to_tail(
        self, key: _torch.Tensor, value: _torch.Tensor, *, layer_idx: int
    ) -> None:
        batch_size, num_new_tokens = key.shape[:2]
        self._validate_input_shapes(batch_size=batch_size, length=num_new_tokens)

        key = key.transpose(0, 1)  # swap batch and sequence dims
        value = value.transpose(0, 1)

        seq_slice = slice(self.length - num_new_tokens, self.length)

        self.cache.keys_and_values[layer_idx, 0, seq_slice] = key
        self.cache.keys_and_values[layer_idx, 1, seq_slice] = value

    def write_to_tail_vectorized(
        self, keys: _torch.Tensor, values: _torch.Tensor, *, layer_indices: _List[int]
    ) -> None:
        batch_size, num_new_tokens = keys.size(0), keys.size(2)
        self._validate_input_shapes(batch_size=batch_size, length=num_new_tokens)

        # input keys and values are in shape [batch, layers, seq, dim]
        keys = keys.permute(1, 2, 0, 3)
        values = values.permute(1, 2, 0, 3)

        if len(layer_indices) != keys.shape[0] or len(layer_indices) != values.shape[0]:
            raise ValueError("Mismatched KV Cache vectorize dimension")

        num_new_tokens = keys.size(1)
        seq_slice = slice(self.length - num_new_tokens, self.length)

        # Vectorized write
        self.cache.keys_and_values[layer_indices, 0, seq_slice] = keys
        self.cache.keys_and_values[layer_indices, 1, seq_slice] = values

    def write_segment_ids_to_tail(self, segment_ids: _torch.Tensor) -> None:
        batch_size, num_new_tokens = segment_ids.shape
        self._validate_input_shapes(batch_size=batch_size, length=num_new_tokens)

        seq_slice = slice(self.length - num_new_tokens, self.length)
        self.cache.segment_ids[seq_slice] = segment_ids.transpose(0, 1)

    def write_positions_to_tail(self, positions: _torch.Tensor) -> None:
        batch_size, num_new_tokens = positions.shape
        self._validate_input_shapes(batch_size=batch_size, length=num_new_tokens)

        seq_slice = slice(self.length - num_new_tokens, self.length)
        self.cache.positions[seq_slice] = positions.transpose(0, 1)

    def write_token_types_to_tail(self, token_types: _torch.Tensor) -> None:
        batch_size, num_new_tokens = token_types.shape
        self._validate_input_shapes(batch_size=batch_size, length=num_new_tokens)

        seq_slice = slice(self.length - num_new_tokens, self.length)
        self.cache.token_types[seq_slice] = token_types.transpose(0, 1)

    def _validate_input_shapes(
        self,
        batch_size: _Optional[int] = None,
        length: _Optional[int] = None,
    ):
        if batch_size is not None and batch_size != self.batch_size:
            raise ValueError(
                f"Input batch size ({batch_size}) differs from the KV cache"
                f"batch size ({self.batch_size})"
            )
        if length is not None and length > self.length:
            raise ValueError(
                f"Input length ({length}) exceeds the KV cache length ({self.length})"
            )

    def read(self, layer_idx: int) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        keys_and_values = self.cache.keys_and_values[layer_idx, :, : self.length]
        keys_and_values = keys_and_values.transpose(1, 2)  # swap batch and seq dims
        return keys_and_values.unbind()

    def read_vectorized(
        self, layer_indices: _List[int]
    ) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        ks = self.cache.keys_and_values[layer_indices, 0, : self.length]
        vs = self.cache.keys_and_values[layer_indices, 1, : self.length]

        # Swap axes -> ( batch, track, ...)
        ks = ks.permute(2, 0, 1, 3)
        vs = vs.permute(2, 0, 1, 3)

        return ks, vs

    def read_segment_ids(self) -> _torch.Tensor:
        segment_ids = self.cache.segment_ids[: self.length]
        return segment_ids.transpose(0, 1)

    def read_positions(self) -> _torch.Tensor:
        positions = self.cache.positions[: self.length]
        return positions.transpose(0, 1)

    def read_token_types(self) -> _torch.Tensor:
        token_types = self.cache.token_types[: self.length]
        return token_types.transpose(0, 1)



import abc as _abc


class CreateFromBuilderMixin(_abc.ABC):
    """
    A mixin for creating and configuring instances of composite modules.  Users have
    three options for creating instances of classes that inherit from this mixin:

    (1) Instantiate the object directly through the :meth:`__init__` method.  Generally
        :meth:`__init__` should accept full child layers (or child builders) as
        arguments, providing the user with maximal flexibility for configuring the
        layer.  While this provides a lot of control, it is often tedious for the user
        to create instances in this way due to complexity of the arguments.
    (2) Create the object by calling the :meth:`create` method.  The :meth:`create`
        method should have a simpler signature and return a "default" instance of the
        class.  This provides much less flexibility compared to option 1, but the
        simpler signature makes it easier to create objects that have common
        configurations.
    (3) Create a builder using the :meth:`create_builder` method, then update the
        builder before using it to create a new instance.  This provides a
        middle-ground between options 1 and 2.  The user provides a simple set of
        arguments to create a builder, and then the user can configure the builder
        arbitrarily before building the composite layer.
    """

    @classmethod
    def create(cls, *args, **kwargs):
        """
        Creates and returns an instance of the class using a builder returned by
        ``cls.create_builder(*args, **kwargs)``.  Please see :meth:`create_builder`
        for a description of supported arguments.

        Args:
            *args: Variable arguments for :meth:`create_builder`.
            *kwargs: Variable keyword arguments for :meth:`create_builder`.
        """
        builder = cls.create_builder(*args, **kwargs)
        return builder()

    @classmethod
    @_abc.abstractmethod
    def create_builder(cls):
        """
        Return a :class:`Builder` for creating instances of the class.  Subclasses
        should add arguments to this method for configuring the builder.
        """



"""
utils.axlearn_utils
===================

This module contains functions for loading and saving AXLearn model checkpoints.

.. autofunction:: tamm.utils.axlearn_utils.load_axlearn_state_dict

.. autofunction:: tamm.utils.axlearn_utils.save_axlearn_state_dict
"""

import collections as _collections
import json as _json
import logging as _logging
import os as _os
import re as _re

import torch as _torch

from tamm import _helpers
from tamm.typing import PathLike as _PathLike
from tamm.typing import StateDictType as _StateDictType

_logger = _logging.getLogger(__name__)

PRNG_KEY = "prng_key"


def load_axlearn_state_dict(
    checkpoint_dir: _PathLike, unstack_repeat_params: bool = True
) -> _StateDictType:
    """
    Loads model state from an AXLearn checkpoint directory.

    Args:
        checkpoint_dir (:obj:`str` or :obj:`PathLike`): The path to the AXLearn checkpoint.
            This should contain an ``index`` JSON file as well as a ``gda`` directory
            that contains parameter values.
        unstack_repeat_params (:obj:`bool`): Flag that when ``True``, transforms the
            result by mapping each parameter with ``"/repeat/"`` in its name to many
            new parameters. This happens by unstacking the parameter along its first
            dimension and by replacing ``"/repeat/"`` in the parameter name with
            ``f"/repeat_{idx}>/"``.

    Returns:
        The loaded model state as a :obj:`dict` that maps parameter names to PyTorch
        tensors.
    """
    _logger.debug("Loading axlearn state dict from %s", checkpoint_dir)
    checkpoint_dir = _get_last_step_dir(checkpoint_dir)
    index_path = _os.path.join(str(checkpoint_dir), "index")
    with open(index_path, "r", encoding="utf-8") as f:
        index = _json.load(f)
    param_paths = [el[0] for el in index if el[0].startswith("model")]
    state_dict = {}
    for param_path in param_paths:
        full_param_path = _os.path.join(str(checkpoint_dir), "gda", param_path)
        ts_tensor = _load_single_axlearn_param(full_param_path)
        tensor = _ts_tensor_to_torch_tensor(ts_tensor)
        state_dict[param_path] = tensor
    if unstack_repeat_params:
        state_dict = _unstack_repeat_params(state_dict)
    _logger.debug("Load complete")
    return state_dict


def _get_last_step_dir(checkpoint_dir):
    """
    AXLearn checkpoint dirs are structured like

    checkpoint_dir/step_00000000/index
    checkpoint_dir/step_00000000/gda/...
    checkpoint_dir/step_00000100/index
    checkpoint_dir/step_00000100/gda/...

    This function infers the path of the step dir with the largest index and returns it
    ("checkpoint_dir/step_100" in the example above).

    If checkpoint_dir already contains index and gda, the function assumes
    checkpoint_dir is the step dir and returns it.
    """
    contents = _os.listdir(checkpoint_dir)
    if "index" in contents and "gda" in contents:
        return checkpoint_dir

    step_indices_and_names = [
        (_get_step_idx_from_basename(name, default=-1), name) for name in contents
    ]
    max_step_idx, max_step_name = max(step_indices_and_names)
    if max_step_idx == -1:
        raise ValueError(
            f"{checkpoint_dir} not recognized as an axlearn checkpoint dir.  Expected it "
            "to contain either (1) a 'step_<step_idx>' subdir or (2) an 'index' file "
            "and 'gda' subdir."
        )
    return _os.path.join(checkpoint_dir, max_step_name)


def _get_step_idx_from_basename(basename: str, default=-1) -> int:
    """
    If basename is of the form "step_{idx}" for an int idx, this returns the idx (as
    an int). Otherwise, this returns ``default``.
    """
    step_idx_regex = _get_step_idx_regex()
    match = step_idx_regex.fullmatch(basename)
    if match is None:
        return default
    return int(match.group(1))


@_helpers.cache
def _get_step_idx_regex():
    return _re.compile(r"step_([\d]+)")


def save_axlearn_state_dict(
    state_dict: _StateDictType,
    checkpoint_dir: _PathLike,
    stack_repeat_params: bool = True,
) -> None:
    """
    Saves a model checkpoint in AXLearn format.

    Args:
        state_dict: A :obj:`dict` that maps parameter names to PyTorch tensors.  The
            names should be relative paths from the ``"gda"`` directory for saving
            each parameter.
        checkpoint_dir (:obj:`str` or :obj:`PathLike`): The path for saving the AXLearn
            checkpoint.
        stack_repeat_params (:obj:`bool`): Flag that when ``True``, transforms the
            result by consolidating parameters with identical names other than the
            layer index in ``"/repeat_<idx>/"``.  The consolidation stacks these
            parameters together and replaces ``"/repeat_<idx>/"`` with ``"/repeat/"``
            in the resulting parameter name.
    """
    _logger.debug("Saving axlearn state dict to %s", checkpoint_dir)
    _os.makedirs(checkpoint_dir, exist_ok=True)
    if stack_repeat_params:
        state_dict = _stack_repeat_params(state_dict)
    checkpoint_dir = _os.path.join(
        str(checkpoint_dir),
        f"step_{0:08d}",
        # Axlearn assumes 8 digits:
        # https://github.com/apple/axlearn/blob/main/axlearn/common/checkpointer.py#L60
    )
    index = [("step", 0)]
    for key, param in _get_state_dict_items(state_dict):
        if _torch.is_tensor(param):  # prng_key is already a numpy array
            param = _torch_tensor_to_ts_tensor(param)
        path = _os.path.join(checkpoint_dir, "gda", key)
        _save_single_axlearn_param(param, path)
        index.append((key, {"dtype": str(param.dtype.name), "shape": str(param.shape)}))
    index_path = _os.path.join(checkpoint_dir, "index")
    with open(index_path, "w", encoding="utf-8") as f:
        _json.dump(index, f)
    _logger.debug("Save complete")


def _torch_tensor_to_ts_tensor(tensor: _torch.Tensor):
    """
    Helper function to convert torch tensors to tensorstore arrays.  To do this, we
    first convert to numpy.  Since numpy does not support some low-precision dtypes, we
    first cast to fp64 and then later back to the original dtype.
    """
    import tensorstore  # pylint: disable=import-outside-toplevel,import-error

    dtype = _helpers.get_str_from_maybe_dtype(tensor.dtype)

    if _torch.is_floating_point(tensor):
        if "mps" in str(tensor.device):
            # Move to CPU because MPS doesn't support double
            tensor = tensor.cpu()
        tensor = tensor.double()
    np_tensor = tensor.numpy(force=True)

    return tensorstore.array(np_tensor).astype(getattr(tensorstore, dtype))


def _ts_tensor_to_torch_tensor(store) -> _torch.Tensor:
    """
    Helper function to convert tensorstores to torch tensors.  We go through numpy, but
    since numpy does not support some low-precision dtypes, we first cast to fp64 and
    later back to the original dtype.
    """
    dtype = store.dtype.name
    torch_dtype = _helpers.get_dtype_from_maybe_string(dtype)
    if "float" in dtype:
        store = store.astype("float64")
    np_tensor = store.read().result()
    return _torch.tensor(np_tensor, dtype=torch_dtype)


def _get_state_dict_items(state_dict):
    """
    Returns an iterable of state dict keys and values.  If prng_key is not already in
    the state_dict, the first item is a prng_key.
    """
    if PRNG_KEY not in state_dict:
        items = [(PRNG_KEY, _create_prng_key_value())]
        items.extend(state_dict.items())
        return items
    return state_dict.items()


def _create_prng_key_value():
    try:
        import numpy as np  # pylint: disable=import-outside-toplevel
    except ImportError:
        _logging.error(
            "Loading axlearn checkpoints requires numpy, but tamm could not import "
            "this package.  Please check your numpy installation."
        )
        raise

    return np.array(
        [1770346568, 1894082449, 1770346568, 1894082449],
        dtype=np.uint32,
        # arbitrary numbers that come from a prior axlearn checkpoint
    )


def _load_single_axlearn_param(path):
    _logger.debug("Loading axlearn param from %s", path)
    return _open_tensorstore(path)


def _save_single_axlearn_param(param, path):
    _logger.debug("Saving axlearn param to %s", path)
    store = _open_tensorstore(path, create=True, dtype=param.dtype, shape=param.shape)
    return store.write(param).result()


def _open_tensorstore(path, **kwargs):
    try:
        import tensorstore  # pylint: disable=import-outside-toplevel
    except ImportError:
        _logging.error(
            "Loading axlearn checkpoints requires tensorstore, but tamm could not import "
            "this package.  Please install tensorstore or check your tensorstore "
            "installation."
        )
        raise

    path = _helpers.get_expanded_abspath(path)  # required by tensorstore
    tensor_store_config = {
        "driver": "zarr",
        "kvstore": {"driver": "file", "path": str(path)},
    }
    return tensorstore.open(tensor_store_config, **kwargs).result()


def _unstack_repeat_params(axlearn_state_dict):
    result = {}
    for key, param in axlearn_state_dict.items():
        if "/repeat/" not in key:
            result[key] = param
            continue
        for idx, new_param in enumerate(param.unbind()):
            new_key = key.replace("/repeat/", f"/repeat_{idx}/")
            result[new_key] = new_param
    return result


def _stack_repeat_params(axlearn_state_dict):
    repeat_id_regex = _get_repeat_id_regex()
    key_to_layer_indices_and_params = _collections.defaultdict(list)
    result = {}
    for key, param in axlearn_state_dict.items():
        match = repeat_id_regex.search(key)
        if match is None:
            result[key] = param
            continue
        new_key = key.replace(match.group(), "/repeat/")
        layer_idx = int(match.group(1))
        key_to_layer_indices_and_params[new_key].append((layer_idx, param))
    for key, value in key_to_layer_indices_and_params.items():
        value.sort()
        indices, params = zip(*value)
        if indices != tuple(range(len(indices))):
            raise ValueError(
                "To stack repeat params, "
                f"expected indices of range({len(indices)}), but indices is {indices}"
            )

        params = [param for _, param in value]
        param = _torch.stack(params)
        result[key] = param
    return result


@_helpers.cache
def _get_repeat_id_regex():
    return _re.compile(r"/repeat_([\d]+)/")



import logging
from pathlib import Path as _Path

from tamm import _helpers
from tamm.utils.uri import _URIHandler

_logger = logging.getLogger(__name__)


def fetch_file(
    remote_path: str,
) -> _Path:
    """
    Fetch large artifacts (e.g., model checkpoints, and tokenizer vocabs) from remote
    storage. Utilizes local diskcache as per implementation selected by URI handler.

    Args:
        remote_path: URI in string

    Returns: Local file path of the first successfully fetched remote path

    """

    return _URIHandler(use_cache=True).map_to_local(remote_path)


def fetch_checkpoint(remote_path: str, *, map_location=None) -> dict:
    """
    Fetch Pytorch checkpoint from remote storage. Utilizes local diskcache as per
    implementation selected by URI handler.

    Args:
        remote_path: URI of remote checkpoint e.g., s3://bucket/ckpt.pt

    Returns: Pytorch state dictionary

    """

    with _URIHandler(use_cache=True).open(remote_path, mode="rb") as fptr:
        state_dict = _helpers.safe_torch_load(fptr, map_location=map_location)
    _logger.info("State dictionary is loaded into memory from the checkpoint file.")
    return state_dict



from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch


def fake_quantize(
    tensor: _torch.Tensor,
    *,
    scale: _torch.Tensor,
    zero_point: _torch.Tensor,
    quant_min: int,
    quant_max: int,
    cast_dtype: _Optional[_torch.dtype] = None,
    reciprocal_mul: bool = False,
):
    """
    Computes fake quantization via the formula

    .. code-block:: python

        quant = (torch.round(tensor / scale) + zero_point).clamp(quant_min, quant_max)
        return scale * (quant - zero_point)

    If ``cast_dtype`` is not ``None``, then the function casts ``tensor`` and ``scale``
    to ``cast_dtype`` before computing the result (but the dtype of the result is the
    same as the input dtype).
    """
    return FakeQuantize.apply(
        tensor, scale, zero_point, quant_min, quant_max, cast_dtype, reciprocal_mul
    )


class FakeQuantize(_torch.autograd.Function):
    # pylint: disable=arguments-differ,abstract-method

    generate_vmap_rule = True

    @staticmethod
    def forward(
        tensor: _torch.Tensor,
        scale: _torch.Tensor,
        zero_point: _torch.Tensor,
        quant_min: int,
        quant_max: int,
        cast_dtype: _Union[_torch.dtype, None],
        reciprocal_mul: bool = False,
    ):
        input_dtype = tensor.dtype
        if reciprocal_mul:
            scale_inv = 1 / scale
        if cast_dtype is not None:
            tensor = tensor.type(cast_dtype)
            if reciprocal_mul:
                scale_inv = scale_inv.type(cast_dtype)
            scale = scale.type(cast_dtype)

        if cast_dtype == input_dtype:
            if reciprocal_mul:
                out = tensor.mul(scale_inv)
            else:
                out = tensor.div(scale)
        else:
            if reciprocal_mul:
                out = tensor.mul_(scale_inv)
            else:
                out = tensor.div_(scale)
        out.round_()
        out.add_(zero_point)
        out.clamp_(quant_min, quant_max)
        out.sub_(zero_point)
        out.mul_(scale)
        out = out.to(input_dtype)
        return out

    @staticmethod
    def setup_context(ctx, inputs, output):
        tensor, *_ = inputs
        ctx.input_dtype = tensor.dtype

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output.type(ctx.input_dtype), *(6 * [None])



import abc
import importlib.metadata
from collections import OrderedDict as _OrderedDict
from typing import Protocol as _Protocol

import packaging.requirements


class JSONDictHook(_Protocol):
    @abc.abstractmethod
    def __call__(self, json_dict: dict) -> dict:
        """
        Classes conform to this protocol should process a json encodable dictionary to
        json encodable dictionary.
        This class intends to define an interface for arbitrary transformation
        needed for json encodable dictionaries
        Args:
            json_dict:

        Returns: a json encodable dict

        """


class ReadabilityHook(JSONDictHook):
    def __init__(self):
        # below is the order that will appear in JSON
        self.attributes_to_bring_forward = [
            "__tamm_type__",
            "__package_requirements__",
            "model_id",
            "description",
            "is_deprecated",
            "replacement_model_id",
        ]
        # flip the order because this operates like a stack
        self.attributes_to_bring_forward.reverse()

    def __call__(self, json_dict: dict):
        json_list = sorted(json_dict.items())
        ordered_dict = _OrderedDict(json_list)
        for key in self.attributes_to_bring_forward:
            try:
                ordered_dict.move_to_end(key, last=False)
            except KeyError:
                pass
        return dict(ordered_dict)


class PackageRequirementsHook(JSONDictHook):
    def __call__(self, json_dict: dict):
        try:
            package_requirements = json_dict.pop("__package_requirements__")
        except KeyError:
            return json_dict
        if package_requirements is not None:
            for requirement_str in package_requirements:
                requirement = packaging.requirements.Requirement(requirement_str)
                package_version = importlib.metadata.version(requirement.name)
                if requirement.specifier.contains(package_version, prereleases=True):
                    continue
                raise RuntimeError(
                    "Attempted to deserialize a tamm object that requires "
                    f"'{requirement_str}', but you have {requirement.name} version "
                    f"{package_version}."
                )
        return json_dict


class UnusedAttributeHook(JSONDictHook):
    def __call__(self, json_dict: dict):
        json_dict.pop("__tamm_type__", None)
        return json_dict



import dataclasses as _dataclasses
from typing import Iterable as _Iterable
from typing import List as _List
from typing import Optional as _Optional
from typing import Union as _Union

from tamm import _helpers
from tamm.utils import partial as _partial
from tamm.utils.json import JSONSerializableMixin as _JSONSerializableMixin


class Tokenizer:
    """
    Base tokenizer class contains basic API for LM tokenization.
    """

    def __len__(self) -> int:
        raise NotImplementedError

    def encode(
        self, texts: _Union[str, _Iterable[str]], **kwargs
    ) -> _Union[_Iterable[int], _Iterable[_Iterable[int]]]:
        raise NotImplementedError

    def decode(
        self,
        token_ids: _Union[int, _Iterable[int], _Iterable[_Iterable[int]]],
        **kwargs,
    ) -> _Union[str, _Iterable[str]]:
        raise NotImplementedError

    def encode_as_pieces(
        self, texts: _Union[str, _Iterable[str]], **kwargs
    ) -> _Union[_List[str], _List[_List[str]]]:
        raise NotImplementedError

    @property
    def eos_id(self) -> int:
        raise NotImplementedError

    @property
    def pad_id(self) -> int:
        raise NotImplementedError

    @property
    def unk_id(self) -> int:
        raise NotImplementedError

    def __init_subclass__(cls):
        tokenizer_name = cls.__name__
        config_name = f"{tokenizer_name}Config"
        config_cls = TokenizerConfig.create_subclass(
            target_callable=cls, name=config_name
        )
        config_cls.__doc__ = (
            f"A :py:class:`.TokenizerConfig` subclass for configuring "
            f":py:class:`.{tokenizer_name}` tokenizer.  Use the alias :attr:`."
            f"{tokenizer_name}.Config` to access this class. "
            f"Please check :class:`.{tokenizer_name}` for more details about the "
            "signature."
        )
        cls.Config = config_cls


class TokenizerConfig(
    _partial.DataclassedPartial,
    _JSONSerializableMixin,
    json_namespace="tokenizers",
):
    # pylint:disable=useless-parent-delegation
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create_tokenizer(self, *override_args, **override_kwargs):
        """
        Creates and returns a configured instance of the tokenizer.

        Args:
            override_args: Optional positional arguments to override args specified in
                the config.  These args replace the first ``len(override_args)``
                positional args.
            override_kwargs: Optional keyword override arguments.  These arguments
                replace any additional named arguments not overriden by
                ``override_args``.

        Returns:
            The newly created tokenizer.
        """
        return self(*override_args, **override_kwargs)

    def _to_json_dict_impl(self):
        result = {}
        for field in _dataclasses.fields(self.configured_args):
            value = getattr(self.configured_args, field.name)
            if value == field.default:
                continue
            result[field.name] = value
        return result


@_dataclasses.dataclass
class PublishedTokenizerConfig(_JSONSerializableMixin, json_namespace="tokenizer_repo"):
    tokenizer_config: TokenizerConfig
    tokenizer_id: int
    is_deprecated: bool = False
    description: _Optional[str] = None

    def _to_json_dict_impl(self):
        return _helpers.dataclass_to_dict(self, omit_defaults=True)

    @classmethod
    def _from_json_dict_impl(cls, **raw_dict: dict):
        return _helpers.dataclass_init_drop_missing_keys(raw_dict, dataclass_type=cls)



import logging as _logging
from typing import List, Union, cast

from tamm._plugin import utils as plugin_utils
from tamm.preprocessors.base import Preprocessor
from tamm.preprocessors.registry import (
    PreprocessorRegistry,
    URIHandlerPreprocessorRegistry,
)
from tamm.runtime_configuration import rc

_logger = _logging.getLogger(__name__)


def get_registry() -> PreprocessorRegistry:
    entrypoints = plugin_utils.discover_plugin_extras(
        rc.PLUGINS_EXTRAS_ENTRYPOINT_NAMES.preprocessor_registry
    )
    preprocessor_registry = cast(PreprocessorRegistry, URIHandlerPreprocessorRegistry())
    for entrypoint in entrypoints:
        extra_registry = plugin_utils.execute_plugin_object_reference(
            entrypoint.value,
            call_obj=True,
        )
        preprocessor_registry += cast(PreprocessorRegistry, extra_registry)

    return preprocessor_registry


def create(
    preprocessor_name: Union[str, "Preprocessor"], *args, **kwargs
) -> "Preprocessor":
    """
    Create a preprocessor instance for the specified preprocessor name.

    Args:
        preprocessor_name (:obj:`str`): The name of the preprocessor to be created.
        *args: Additional positional arguments for preprocessor creation.
        **kwargs: Additional keyword arguments for preprocessor creation.

    Returns:
        :obj:`Preprocessor`: Created preprocessor instance.
    """
    return get_registry().create(preprocessor_name, *args, **kwargs)


def describe(preprocessor_name: Union[str, "Preprocessor"], *args, **kwargs) -> dict:
    """
    Describe a preprocessor instance by a dictionary

    Args:
        preprocessor_name (:obj:`str`): The name of the preprocessor to be created.
        *args: Additional positional arguments for preprocessor creation.
        **kwargs: Additional keyword arguments for preprocessor creation.

    Returns:
        :obj:`dict`: Dictionary which describes a preprocessor.
    """
    return get_registry().describe(preprocessor_name, *args, **kwargs)


def list_objects() -> List[str]:
    """
    List available preprocessors


    Returns:
        Union[List[str], Dict[str, str]]: A list of preprocessors IDs if

    """
    return get_registry().list_objects()



"""
tamm.converters
---------------

This module implements tools to convert |tamm| state dicts to and from other formats.

Base class
^^^^^^^^^^

.. autoclass:: tamm.converters.StateDictConverter
    :members:


Building blocks
^^^^^^^^^^^^^^^

.. autoclass:: tamm.converters.UnionConverter
    :show-inheritance:

.. autoclass:: tamm.converters.SequentialConverter
    :show-inheritance:

.. autoclass:: tamm.converters.LayerwiseConverter
    :show-inheritance:

.. autoclass:: tamm.converters.ManyToManyStringMatchingConverter
    :show-inheritance:

.. autoclass:: tamm.converters.MultiPrefixStringMatchingConverter
    :show-inheritance:

.. autoclass:: tamm.converters.ParamMapper
    :show-inheritance:

.. autoclass:: tamm.converters.PrefixStringMatchingConverter
    :show-inheritance:

.. autoclass:: tamm.converters.StringMatchingConverter
    :show-inheritance:
"""

from tamm.converters import layers
from tamm.converters.common import (
    LayerwiseConverter,
    ManyToManyStringMatchingConverter,
    MultiPrefixStringMatchingConverter,
    ParamMapper,
    PrefixStringMatchingConverter,
    SequentialConverter,
    StateDictConverter,
    StringMatchingConverter,
    UnionConverter,
)
from tamm.converters.models import (
    ModelStateDictConverter,
    convert_from_tamm_state_dict,
    convert_to_tamm_state_dict,
    list_converters,
    load,
    save,
)



"""
transformer.token_metadata
^^^^^^^^^^^^^^^^^^^^^^^^^^

This module defines classes for handling token metadata within a
:class:`.TransformerStack`.


.. autoclass:: tamm.layers.transformer.token_metadata.TokenMetadataLogic
    :show-inheritance:
    :members: forward

.. autoclass:: tamm.layers.transformer.token_metadata.TokenMetadata
    :members:

.. autoclass:: tamm.layers.transformer.token_metadata.VanillaTokenMetadataLogic
    :show-inheritance:
"""

from tamm.layers.transformer.token_metadata.common import (
    TokenMetadata,
    TokenMetadataLogic,
)
from tamm.layers.transformer.token_metadata.vanilla import VanillaTokenMetadataLogic



"""HF transformers model configuration for tamm wrappers"""

import json as _json
from typing import Optional as _Optional

from transformers import configuration_utils as _configuration_utils

import tamm as _tamm


class TammPretrainedConfig(_configuration_utils.PretrainedConfig):
    def __init__(
        self, tamm_config: _Optional[_tamm.models.common.ModelConfig] = None, **kwargs
    ):
        """
        HuggingFace transformers config for wrapper |tamm| models.

        Args:
            tamm_config (:obj:`ModelConfig`): The config for the |tamm| model.
            **kwargs: Keyword arguments for Hugging Face default options.
        """
        super().__init__(**kwargs)
        self.tamm_config = tamm_config

    def to_dict(self):
        return self._to_dict_impl(omit_defaults=False)

    def to_diff_dict(self):
        return self._to_dict_impl(omit_defaults=True)

    def _to_dict_impl(self, omit_defaults):
        tamm_config = self.tamm_config
        self.tamm_config = None

        result = super().to_diff_dict() if omit_defaults else super().to_dict()

        tamm_json = _tamm.utils.json.dumps(tamm_config)
        tamm_dict = _json.loads(tamm_json)
        result["tamm_config"] = tamm_dict

        self.tamm_config = tamm_config

        return result

    @classmethod
    def from_dict(cls, config_dict, **kwargs):
        tamm_dict = config_dict.get("tamm_config")
        if tamm_dict is not None:
            config_dict = config_dict.copy()
            tamm_json = _json.dumps(tamm_dict)
            tamm_config = _tamm.utils.json.loads(tamm_json)
            config_dict["tamm_config"] = tamm_config
        return super().from_dict(config_dict, **kwargs)

    @classmethod
    def from_json_file(cls, json_file):
        with open(json_file, "r", encoding="utf-8") as reader:
            text = reader.read()
        config_dict = _json.loads(text)
        return cls.from_dict(config_dict)



import functools as _functools
import json as _json
import warnings as _warnings
from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional

from tamm import tokenizers as _tokenizers
from tamm.preprocessors import base as _base
from tamm.preprocessors.instruct_lm import common as _common
from tamm.typing import LenientOptionalBool as _LenientOptionalBool
from tamm.utils import OptionalBool as _OptionalBool

DEFAULT_SYSTEM_MESSAGE = "A conversation between a user and a helpful assistant."
SYSTEM_ROLE = "system"


class JinjaChatTemplatePreprocessor(_base.Preprocessor):
    """
    An instruct LM preprocessor that uses Jinja for chat templating,
    similar to Hugging Face tokenizers.

    Args:
        tokenizer_spec: An argument to :func:`tamm.create_tokenizer` for
            creating a text tokenizer.
        chat_template (:obj:`str`): A Jinja template for preparing inputs
            to the tokenizer.  The template can accept ``messages`` and
            ``tools`` passed to :meth:`__call__` as well as an
            ``add_generation_prompt`` flag and ``target_role`` string.
        generate_prefix (:obj:`bool`): A flag for appending tokens to
            indicate the start of a response.  Typically this value is
            ``False`` during training and ``True`` during inference.
            Defaults to ``False``.
        max_sequence_length (:obj:`int`, optional): An optional max length
            for the preprocessed sequence.  If not ``None``, the
            preprocessor drops any tokens beyond this length.  This only
            applies when ``generate_prefix`` is ``False``. Defaults to
            ``None``.
        prepend_system_message (:obj:`bool`): A flag for prepending a system
            message to the input messages.  Set to ``False`` to pass a system
            message when calling the preprocessor.  Defaults to ``True``.
        system_message (:obj:`str`): A system message for providing the
            assistant with context or instructions for the conversation.
            Defaults to
            ``"A conversation between a user and a helpful assistant."``.
        strip_first_token (:obj:`bool`): A flag for stripping the first
            token from the output of the tokenizer.  Defaults to ``False``.
        prepend_eos (:obj:`bool`): A flag for prepending an EOS token id
            to the result.  Defaults to ``False``.
        target_role (:obj:`str`): The role name of the assistant.  Defaults
            to ``"assistant"``.
        ignored_id (:obj:`int`): The label ID for tokens that should be
            ignored during loss calculation.  Defaults to ``-100``.
    """

    def __init__(
        self,
        tokenizer_spec: _tokenizers.TokenizerSpecType,
        chat_template: str,
        generate_prefix: bool = False,
        max_sequence_length: _Optional[int] = None,
        prepend_system_message: bool = True,
        system_message: str = DEFAULT_SYSTEM_MESSAGE,
        strip_first_token: bool = False,
        prepend_eos: bool = False,
        target_role: str = "assistant",
        ignored_id: int = -100,
    ):
        self.tokenizer_spec = tokenizer_spec
        self.chat_template = chat_template
        self.generate_prefix = generate_prefix
        self.max_sequence_length = max_sequence_length

        self.prepend_system_message = prepend_system_message
        if system_message is None:
            system_message = DEFAULT_SYSTEM_MESSAGE
        self.system_message = system_message

        self.prepend_eos = prepend_eos
        self.strip_first_token = strip_first_token

        self.target_role = target_role
        self.ignored_id = ignored_id

    @_functools.cached_property
    def tokenizer(self) -> _tokenizers.common.Tokenizer:
        """The tokenizer used by the preprocessor."""
        return _tokenizers.create_tokenizer(self.tokenizer_spec)

    @_functools.cached_property
    def _jinja_chat_template(self):
        # pylint: disable-next=import-outside-toplevel
        from jinja2 import sandbox as _jinja2_sandbox

        env = _jinja2_sandbox.ImmutableSandboxedEnvironment(
            trim_blocks=True, lstrip_blocks=True
        )
        env.filters["tojson"] = self._jinja_to_json
        return env.from_string(self.chat_template)

    @staticmethod
    def _jinja_to_json(value, indent=None):
        return _json.dumps(value, indent=indent, ensure_ascii=False)

    @_functools.cached_property
    def _target_role_generation_prefix(self) -> _List[int]:
        """
        Returns a list of token ids that signifies the start of a message
        from the target role (assistant).
        """
        messages = [
            {"role": SYSTEM_ROLE, "content": "A conversation"},
        ]
        if self.target_role != "user":
            messages.append({"role": "user", "content": "Hello"})

        chat_no_generation_prefix = self._jinja_chat_template.render(messages=messages)
        tokens_no_generation_prefix = self.tokenizer.encode(chat_no_generation_prefix)

        chat_with_generation_prefix = self._jinja_chat_template.render(
            messages=messages,
            target_role=self.target_role,
            add_generation_prompt=True,
        )
        tokens_with_generation_prefix = self.tokenizer.encode(
            chat_with_generation_prefix
        )
        return tokens_with_generation_prefix[len(tokens_no_generation_prefix) :]

    def __call__(
        self,
        messages: _List[_Dict[str, str]],
        *,
        generate_prefix: _LenientOptionalBool = _OptionalBool.NOTSET,
        tools: _Optional[_List[_Dict[str, _Any]]] = None,
    ):
        """
        Preprocesses a list of raw messages.

        Args:
            messages (:obj:`list`): The list of messages to pass to
                the Jinja template.  Each message must be a :obj:`dict`
                that contains at minimum ``"role"`` and ``"content"``
                keys.
            tools (:obj:`list`): A list of tool definitions to pass
                to the Jinja template.
            generate_prefix (:obj:`bool`, optional): An optional
                flag for overriding the preprocessor's
                ``generate_prefix`` option.  Typically ``True`` for
                inference and ``False`` for training.

        Returns: A :obj:`InstructLMPreprocessorOutput` that contains the
            tokenized ``input_ids`` and also ``label_ids``.  The labels
            are ``None`` when the input ``messages`` do not contain
            content from the preprocessor's target role.
            Otherwise the labels contain content from the target role
            for training or evaluation.
        """

        if len(messages) == 0:
            raise ValueError("Input must contain at least one message")
        if generate_prefix is _OptionalBool.NOTSET:
            generate_prefix = self.generate_prefix

        messages = self.maybe_prepend_system_message(messages)
        messages = [self.maybe_update_message(message) for message in messages]

        if tools is not None and messages[0]["role"] != SYSTEM_ROLE:
            raise ValueError(
                "When using tools, the first message must be a system "
                f"message but instead it is {messages[0]}"
            )
        if not generate_prefix:
            self._validate_messages_contain_target_role(messages)

        text = self._jinja_chat_template.render(
            messages=messages,
            target_role=self.target_role,
            tools=tools,
            add_generation_prompt=generate_prefix,
        )

        token_ids = self.tokenizer.encode(text)
        if self.strip_first_token:
            token_ids = token_ids[1:]
        if self.prepend_eos:
            token_ids = [self.tokenizer.eos_id] + token_ids

        input_ids, label_ids = self._create_input_and_label_ids_from_token_ids(
            token_ids=token_ids,
            messages=messages,
            generate_prefix=generate_prefix,
        )

        return _common.InstructLMPreprocessorOutput(
            input_ids=input_ids, label_ids=label_ids
        )

    def maybe_prepend_system_message(self, messages):
        if any(message["role"] == SYSTEM_ROLE for message in messages):
            if messages[0]["role"] != SYSTEM_ROLE:
                raise ValueError(f"System is not the first role in messages {messages}")
            if self.prepend_system_message:
                raise ValueError(
                    "prepend_system_message is True, but messages already contains a system "
                    f"message: {messages}"
                )
        elif self.prepend_system_message:
            if len(self.system_message) == 0:
                _warnings.warn(
                    "System message is empty; please set prepend_system_message=False to avoid "
                    "prepending a system message"
                )
            elif self.system_message[0] == " " or self.system_message[-1] == " ":
                raise ValueError(
                    "Custom system message should not start or end with a space but "
                    f"system_message='{self.system_message}'"
                )
            messages = [
                {"role": SYSTEM_ROLE, "content": self.system_message}
            ] + messages
        else:
            _warnings.warn(
                "No system message found but not prepending a system message because "
                "prepend_system_message is False"
            )
        return messages

    def maybe_update_message(self, message: _Dict[str, _Any]) -> _Dict[str, _Any]:
        """
        A helper function that provides subclasses a way to preprocess messages
        in Python prior to Jinja template rendering.
        """
        return message

    def _validate_messages_contain_target_role(self, messages):
        if any(self.target_role == msg["role"] for msg in messages):
            return
        raise ValueError(
            f"Target role '{self.target_role}' is missing from messages "
            f"(if the preprocessing is for {self.target_role} generation, please "
            "set generate_prefix=True)"
        )

    def _create_input_and_label_ids_from_token_ids(
        self, *, token_ids, messages, generate_prefix
    ):
        # pylint: disable=too-many-locals
        if generate_prefix and messages[-1]["role"] != self.target_role:
            # In this case (inference), there are no labels
            return token_ids, None

        generation_prefix = self._target_role_generation_prefix
        generation_prefix = "".join(f",{i}" for i in generation_prefix)

        if generate_prefix:
            # In this case (inference), the last message serves as the labels
            str_token_ids = "".join(f",{i}" for i in token_ids)
            eot_and_prefix = "," + str(self.tokenizer.eot_id) + generation_prefix
            str_labels_ids = str_token_ids.rsplit(eot_and_prefix, maxsplit=1)[-1]
            num_labels = str_labels_ids.count(",")
            return token_ids[:-num_labels], token_ids[-num_labels:]

        label_ids = []
        start_idx = 0
        for message in messages:
            try:
                end_idx = token_ids.index(self.tokenizer.eot_id, start_idx) + 1
            except ValueError:
                end_idx = len(token_ids)

            if message["role"] != self.target_role:
                # Use ignore label ids for non-target role
                label_ids.extend([self.ignored_id] * (end_idx - start_idx))
            else:
                # Extract content as labels for target role
                subsequence = token_ids[start_idx:end_idx]
                str_subsequence = "".join(f",{i}" for i in subsequence)
                labels_part = str_subsequence.split(generation_prefix, maxsplit=1)[-1]
                num_labels = labels_part.count(",")
                num_ignore = len(subsequence) - num_labels
                label_ids.extend([self.ignored_id] * num_ignore)
                label_ids.extend(subsequence[-num_labels:])

            start_idx = end_idx

        # Shift labels
        label_ids.append(self.ignored_id)
        label_ids = label_ids[1:]

        if self.max_sequence_length is not None:
            token_ids = token_ids[: self.max_sequence_length]
            label_ids = label_ids[: self.max_sequence_length]

        return token_ids, label_ids



import abc as _abc
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch

from tamm.layers import side_outputs as _side_outputs
from tamm.layers.transformer import kv_cache as _kv_cache
from tamm.layers.transformer import token_metadata as _token_metadata


class BaseTransformerLayerSequence(_torch.nn.Module, _abc.ABC):
    """
    A base class that defines the interface for transformer layer sequences.
    """

    def __init__(self):
        super().__init__()

    @property
    @_abc.abstractmethod
    def num_layers(self):
        """The number of layers within the layer sequence."""

    @_abc.abstractmethod
    def forward(
        self,
        hidden_states: _torch.Tensor,
        *,
        attention_side_inputs: _Dict[str, _Any],
        kv_cache: _Optional[_kv_cache.BaseKVCacheView] = None,
        secondary_attention_side_inputs: _Optional[_Dict[str, _Dict[str, _Any]]] = None,
        cross_attention_side_inputs: _Dict[str, _Any],
        token_metadata: _token_metadata.TokenMetadata,
    ) -> _Union[_torch.Tensor, _side_outputs.OutputWithSideOutputs]:
        """
        Args:
            hidden_states (:obj:`torch.Tensor`): Input hidden states to the transformer
                layers.
            attention_side_inputs (:obj:`dict`): Keyword arguments to self-attention
                layers within the layer sequence, excluding KV cache values.
            kv_cache (:obj:`.BaseKVCacheView`, optional): Optional cached keys
                and values for self-attention layers.
            secondary_attention_side_inputs (:obj:`dict`, optional): The output of a
                :class:`.SecondaryPositionalEncodings` layer, which contains
                additional attention side inputs for different types of positional
                encodings.
            cross_attention_side_inputs (:obj:`dict`): Keyword arguments to
                cross-attention layers within the sequence.
            token_metadata (:obj:`TokenMetadata`): The :obj:`TokenMetadata` from
                the :class:`TransformerStack`.
        """



import json as _json
import logging as _logging
from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List

_logger = _logging.getLogger(__name__)


def tool_calls_to_json_str(tool_calls: _List[_Dict[str, _Any]]) -> str:
    tool_calls = [
        {
            "name": call["function"]["name"],
            "arguments": _parse_tool_call_arguments(call["function"]["arguments"]),
        }
        for call in tool_calls
    ]
    return _json.dumps(tool_calls, ensure_ascii=False)


def _parse_tool_call_arguments(arguments: str) -> _Any:
    try:
        return _json.loads(arguments)
    except _json.JSONDecodeError:
        _logger.debug("Failed to parse tool call arguments %s", arguments)
    return arguments


def tools_to_json_str(tools: _List[_Dict[str, _Any]]) -> str:
    tools = [_json.dumps(tool, ensure_ascii=False) for tool in tools]
    return "\n".join(tools)



"""
transformer.text
^^^^^^^^^^^^^^^^

This submodule implements transformer-specific components for language modeling.

.. autoclass:: tamm.layers.CausalLMTransformer
    :show-inheritance:
    :members: forward

.. autoclass:: tamm.layers.transformer.CausalLMTransformerOutput
    :members:
    :inherited-members:

.. autoclass:: tamm.layers.transformer.TextTransformerEncoder
"""
import collections as _collections
from contextlib import contextmanager as _contextmanager
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch

from tamm import _helpers
from tamm.layers import linear as _linear
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.transformer import kv_cache as _kv_cache
from tamm.layers.transformer import stack as _transformer_stack
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder
from tamm.typing import OptionalTensor as _OptionalTensor
from tamm.utils import torch_utils as _torch_utils


class CausalLMTransformer(_transformer_stack.ExtendedTransformerStack, _LayerMixin):
    """
    A general layer for decoder-only causal language models.  This layer is the same as
    :class:`.ExtendedTransformerStack` except for a few changes:

    * The layer has an optional ``segmentation`` layer, which computes default
      ``segment_ids`` when the user passes token ids but no segment IDs.
    * The :meth:`forward` method does not take arguments for cross attention.

    Args:
        segmentation: A layer for converting input token IDs to segment IDs.  Often this
            is a padding mask that returns ``0`` for padding tokens and ``1``
            for non-padding.

    Please see :class:`.ExtendedTransformerStack` for a description of the remaining
    parameters.
    """

    def __init__(
        self,
        *,
        segmentation: _OptionalModuleOrBuilder = None,
        embedding: _OptionalModuleOrBuilder = None,
        token_metadata_logic: _OptionalModuleOrBuilder = None,
        token_type_encoding: _OptionalModuleOrBuilder = None,
        attention_mask: _OptionalModuleOrBuilder = None,
        positional_encoding: _OptionalModuleOrBuilder = None,
        secondary_positional_encodings: _OptionalModuleOrBuilder = None,
        input_norm: _OptionalModuleOrBuilder = None,
        layers: _ModuleOrBuilder,
        output_norm: _OptionalModuleOrBuilder = None,
        output_transform: _OptionalModuleOrBuilder = None,
        output_transform_build_mode: _Optional[str] = "basic",
        loss: _OptionalModuleOrBuilder = None,
        embeddings_cast_dtype: _Optional[_Union[str, _torch.dtype]] = "auto",
    ):
        super().__init__(
            embedding=embedding,
            token_metadata_logic=token_metadata_logic,
            token_type_encoding=token_type_encoding,
            attention_mask=attention_mask,
            positional_encoding=positional_encoding,
            secondary_positional_encodings=secondary_positional_encodings,
            input_norm=input_norm,
            layers=layers,
            output_norm=output_norm,
            output_transform=output_transform,
            output_transform_build_mode=output_transform_build_mode,
            loss=loss,
            embeddings_cast_dtype=embeddings_cast_dtype,
        )
        _helpers.prepend_children(self, segmentation=segmentation)
        # Flag to enable tracing behavior, when set to `True`, returns dict instead of dataclass.
        self._return_tuple_output = False

    def forward(  # pylint: disable=arguments-differ
        self,
        inputs: _Any,
        *,
        labels: _Any = None,
        segment_ids: _OptionalTensor = None,
        positions: _OptionalTensor = None,
        token_types: _OptionalTensor = None,
        kv_cache: _Optional[_kv_cache.BaseKVCacheView] = None,
        attention_mask: _OptionalTensor = None,
        cross_attention_source: _OptionalTensor = None,
        cross_attention_mask: _OptionalTensor = None,
        aux_token_metadata: _Optional[_Dict[str, _Any]] = None,
        mode: _Union[_transformer_stack.ExtendedTransformerStackMode, str] = "full",
    ):
        """
        Args:
            inputs (:obj:`torch.Tensor`): Inputs to the ``embedding`` layer, which
                typically are input token IDs with shape ``(batch_size, seq_len)``.  For
                some values of ``mode``, this should instead be input embeddings with
                shape ``(batch_size, seq_len, hidden_dim)``.
            labels: Optional labels for the ``loss`` layer.  If provided, the loss layer
                should accept two positional arguments; the first is the output of
                ``output_transform``, and the second is ``labels``.
            segment_ids (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that controls the attention mask for
                self attention layers.  For sequence ``i``, if
                ``segment_ids[i, j] != segment_ids[i, k]``, then tokens ``j`` and
                ``k`` do not attend to each other.  If unspecified, the ``kv_cache``
                argument (if provided) or ``segmentation`` layer (otherwise) determines
                the ``segment_ids``.  If neither of these are provided, ``segment_ids``
                defaults to all ones.
            positions (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that specifies the position of each
                token.  If unspecified, ``positions`` defaults to
                ``0, 1, 2, ...`` for each segment of each sequence.
            token_types (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that specifies a "type" for each token.
                The ``token_type_encoding`` and ``attention_mask`` child layers may use
                this type information in different ways, depending on the model.
            kv_cache (:obj:`BaseKVCacheView`, optional): A KV cache view with cached
                keys and values for self-attention layers.
            attention_mask (:obj:`torch.Tensor`, optional): A boolean tensor with shape
                ``(batch_size, seq_len, kv_seq_len)``.  A ``True`` value indicates that
                the corresponding pair of tokens should take part in attention.  The
                ``kv_seq_len`` should be the same as ``seq_len`` unless using a
                ``kv_cache``, in which case ``kv_seq_len`` is the cache length.
            cross_attention_source (:obj:`torch.Tensor`, optional): Input source
                embeddings for possible cross attention layers with shape
                ``(batch_size, source_seq_len, hidden_dim)``. Note: causal LMs typically
                do not have cross attention layers, but this is sometimes useful for
                extending the model (for example to handle long contexts).
            cross_attention_mask (:obj:`torch.Tensor`, optional): An attention mask for
                cross attention with shape ``(batch_size, seq_len, source_seq_len)``.
            aux_token_metadata (:obj:`dict`): A dictionary of optional auxiliary inputs for
                extending model behavior.  This gets attached to the ``token_metadata``
                input of several child layers, including the attention mask, layer
                sequence, and positional encoding layers.
            mode (:obj:`str` or :obj:`.ExtendedTransformerStackMode`, optional): A
                mode for controlling the forward behavior.  Defaults to ``FULL``.
        """

        mode = _helpers.get_enum_member_from_name(
            _transformer_stack.ExtendedTransformerStackMode, mode
        )
        segment_ids = self._maybe_update_segment_ids(
            segment_ids, inputs=inputs, mode=mode, kv_cache=kv_cache
        )
        # pylint: disable=duplicate-code
        output = super().forward(
            inputs=inputs,
            labels=labels,
            segment_ids=segment_ids,
            positions=positions,
            token_types=token_types,
            kv_cache=kv_cache,
            attention_mask=attention_mask,
            cross_attention_source=cross_attention_source,
            cross_attention_mask=cross_attention_mask,
            aux_token_metadata=aux_token_metadata,
            mode=mode,
        )
        output_dict = _helpers.dataclass_to_dict(output)

        if self._return_tuple_output:
            return self._create_output_tuple(output_dict)

        return CausalLMTransformerOutput(**output_dict)

    @staticmethod
    def _create_output_tuple(output_dict):
        """Creates a named tuple from the output dictionary, filtering None values."""
        output_dict_without_none = {
            k: v for k, v in output_dict.items() if v is not None
        }
        output_type = _collections.namedtuple(
            "CausalLMDynamicOutput", list(output_dict_without_none.keys())
        )
        return output_type(**output_dict_without_none)

    def _maybe_update_segment_ids(self, segment_ids, *, inputs, mode, kv_cache):
        if segment_ids is not None:
            return segment_ids
        if self._should_skip_embedding_layer(mode):
            return None
        if self.segmentation is None:
            return None
        if kv_cache is not None:
            return None
        return self.segmentation(inputs)

    @property
    def is_embedding_tied_to_output_transform(self):
        if not isinstance(self.output_transform, _linear.TiedWeightLinear):
            return False
        try:
            return self.embedding.weight is self.output_transform.weight
        except AttributeError:
            return False

    @_contextmanager
    def jit_traceable_context(self):
        """
        Context manager to enable torch.jit.trace by returning tuple outputs
        instead of dataclass during tracing.
        """
        original_output = self._return_tuple_output
        self._return_tuple_output = True

        try:
            yield
        finally:
            self._return_tuple_output = original_output


@_torch_utils.torch_exportable_dataclass
class CausalLMTransformerOutput(_transformer_stack.ExtendedTransformerStackOutput):
    """A dataclass for holding outputs from a :class:`CausalLMTransformer`."""

    @property
    def logits(self):
        """An alias of :attr:`predictions`."""
        return self.predictions


class TextTransformerEncoder(_transformer_stack.ExtendedTransformerStack, _LayerMixin):
    """
    A general layer for text encoder language models.  This layer is the same as
    :class:`.ExtendedTransformerStack` except for a few changes:

    * The layer has an optional ``segmentation`` layer, which computes default
      ``segment_ids`` when the user passes token ids but no segment IDs.
    * The :meth:`forward` method does not take arguments for cross attention.

    Args:
        segmentation: A layer for converting input token IDs to segment IDs.  Often this
            is a padding mask that returns ``0`` for padding tokens and ``1``
            for non-padding.

    Please see :class:`.ExtendedTransformerStack` for a description of the remaining
    parameters.
    """

    def __init__(
        self,
        *,
        segmentation: _OptionalModuleOrBuilder = None,
        embedding: _OptionalModuleOrBuilder = None,
        token_metadata_logic: _OptionalModuleOrBuilder = None,
        token_type_encoding: _OptionalModuleOrBuilder = None,
        attention_mask: _OptionalModuleOrBuilder = None,
        positional_encoding: _OptionalModuleOrBuilder = None,
        secondary_positional_encodings: _OptionalModuleOrBuilder = None,
        input_transform: _OptionalModuleOrBuilder = None,
        input_norm: _OptionalModuleOrBuilder = None,
        layers: _ModuleOrBuilder,
        output_norm: _OptionalModuleOrBuilder = None,
        output_transform: _OptionalModuleOrBuilder = None,
        output_transform_build_mode: _Optional[str] = "basic",
        loss: _OptionalModuleOrBuilder = None,
        embeddings_cast_dtype: _Optional[_Union[str, _torch.dtype]] = "auto",
    ):
        # pylint: disable=too-many-locals
        super().__init__(
            embedding=embedding,
            token_metadata_logic=token_metadata_logic,
            token_type_encoding=token_type_encoding,
            attention_mask=attention_mask,
            positional_encoding=positional_encoding,
            secondary_positional_encodings=secondary_positional_encodings,
            input_transform=input_transform,
            input_norm=input_norm,
            layers=layers,
            output_norm=output_norm,
            output_transform=output_transform,
            output_transform_build_mode=output_transform_build_mode,
            loss=loss,
            embeddings_cast_dtype=embeddings_cast_dtype,
        )
        _helpers.prepend_children(self, segmentation=segmentation)

    def forward(  # pylint: disable=arguments-differ
        self,
        inputs: _Any,
        *,
        labels: _Any = None,
        segment_ids: _OptionalTensor = None,
        positions: _OptionalTensor = None,
        token_types: _OptionalTensor = None,
        kv_cache: _Optional[_kv_cache.BaseKVCacheView] = None,
        attention_mask: _OptionalTensor = None,
        aux_token_metadata: _Optional[_Dict[str, _Any]] = None,
        mode: _Union[_transformer_stack.ExtendedTransformerStackMode, str] = "full",
    ) -> "_transformer_stack.ExtendedTransformerStackOutput":
        """
        Args:
            inputs (:obj:`torch.Tensor`): Inputs to the ``embedding`` layer, which
                typically are input token IDs with shape ``(batch_size, seq_len)``.  For
                some values of ``mode``, this should instead be input embeddings with
                shape ``(batch_size, seq_len, hidden_dim)``.
            labels: Optional labels for the ``loss`` layer.  If provided, the loss layer
                should accept two positional arguments; the first is the output of
                ``output_transform``, and the second is ``labels``.
            segment_ids (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that controls the attention mask for
                self attention layers.  For sequence ``i``, if
                ``segment_ids[i, j] != segment_ids[i, k]``, then tokens ``j`` and
                ``k`` do not attend to each other.  If unspecified, the ``kv_cache``
                argument (if provided) or ``segmentation`` layer (otherwise) determines
                the ``segment_ids``.  If neither of these are provided, ``segment_ids``
                defaults to all ones.
            positions (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that specifies the position of each
                token.  If unspecified, ``positions`` defaults to
                ``0, 1, 2, ...`` for each segment of each sequence.
            token_types (:obj:`torch.Tensor`, optional): An integer tensor with shape
                ``(batch_size, seq_len)`` that specifies a "type" for each token.
                The ``token_type_encoding`` and ``attention_mask`` child layers may use
                this type information in different ways, depending on the model.
            kv_cache (:obj:`BaseKVCacheView`, optional): A KV cache view with cached
                keys and values for self-attention layers.
            attention_mask (:obj:`torch.Tensor`, optional): A boolean tensor with shape
                ``(batch_size, seq_len, kv_seq_len)``.  A ``True`` value indicates that
                the corresponding pair of tokens should take part in attention.  The
                ``kv_seq_len`` should be the same as ``seq_len`` unless using a
                ``kv_cache``, in which case ``kv_seq_len`` is the cache length.
            aux_token_metadata (:obj:`dict`): A dictionary of optional auxiliary inputs for
                extending model behavior.  This gets attached to the ``token_metadata``
                input of several child layers, including the attention mask, layer
                sequence, and positional encoding layers.
            mode (:obj:`str` or :obj:`.ExtendedTransformerStackMode`, optional): A
                mode for controlling the forward behavior.  Defaults to ``FULL``.
        """

        mode = _helpers.get_enum_member_from_name(
            _transformer_stack.ExtendedTransformerStackMode, mode
        )
        segment_ids = self._maybe_update_segment_ids(
            segment_ids, inputs=inputs, mode=mode, kv_cache=kv_cache
        )
        output = super().forward(
            inputs=inputs,
            labels=labels,
            segment_ids=segment_ids,
            positions=positions,
            token_types=token_types,
            kv_cache=kv_cache,
            attention_mask=attention_mask,
            aux_token_metadata=aux_token_metadata,
            mode=mode,
        )
        return _transformer_stack.ExtendedTransformerStackOutput(
            **_helpers.dataclass_to_dict(output)
        )

    def _maybe_update_segment_ids(self, segment_ids, *, inputs, mode, kv_cache):
        if segment_ids is not None:
            return segment_ids
        if self._should_skip_embedding_layer(mode):
            return None
        if self.segmentation is None:
            return None
        if kv_cache is not None:
            return None
        return self.segmentation(inputs)



"""
Builder Context Hooks
---------------------

.. autoclass:: tamm.layers.common.DefaultDeviceContextHook
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.common.DtypeContextHook
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.common.PretrainedContextHook
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.common.UseMetaInitTrickContextHook
    :members:
    :show-inheritance:
"""

from tamm.layers.common.context_hooks.default_device import DefaultDeviceContextHook
from tamm.layers.common.context_hooks.dtype import DtypeContextHook
from tamm.layers.common.context_hooks.freeze_params import FreezeParamsContextHook
from tamm.layers.common.context_hooks.pretrained import PretrainedContextHook
from tamm.layers.common.context_hooks.use_meta_init_trick import (
    UseMetaInitTrickContextHook,
)

__all__ = [
    "DefaultDeviceContextHook",
    "DtypeContextHook",
    "PretrainedContextHook",
    "UseMetaInitTrickContextHook",
    "FreezeParamsContextHook",
]



"""
.. _adapter_api:

adapters.adapter_api
====================

.. autofunction:: tamm.adapters.init

.. autofunction:: tamm.adapters.is_adapter_initialized

.. autofunction:: tamm.adapters.has_adapter

.. autofunction:: tamm.adapters.delete_adapter

.. autofunction:: tamm.adapters.freeze_adapter

.. autofunction:: tamm.adapters.unfreeze_adapter

.. autofunction:: tamm.adapters.set_active_adapter

.. autofunction:: tamm.adapters.unset_active_adapter

.. autofunction:: tamm.adapters.merge_adapter

.. autofunction:: tamm.adapters.get_all_adapter_ids

.. autofunction:: tamm.adapters.get_all_active_adapter_ids

.. autofunction:: tamm.adapters.map_state_dict

.. autofunction:: tamm.adapters.load_state_dict

.. autoclass:: tamm.adapters.AdaptedModelStateDictFormat
"""

import enum as _enum
import logging as _logging
import warnings as _warnings
from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Set as _Set

import torch.nn as _nn
from torch.nn.modules.module import _IncompatibleKeys

from tamm import _helpers
from tamm._adapters_v1.adapted_layer import AdaptedLayer as _AdaptedLayer
from tamm._adapters_v1.layer_adapters.common import (
    AdapterWithExtraInputs as _AdapterWithExtraInputs,
)
from tamm._adapters_v1.layer_annotations import (
    is_adaptable_layer as _is_adaptable_layer,
)
from tamm._adapters_v1.shared_context import SharedContext as _SharedContext
from tamm._adapters_v1.utils import is_v0_lora_state_dict as _is_v0_lora_state_dict
from tamm._adapters_v1.utils import maybe_rekey_state_dict as _maybe_rekey_state_dict
from tamm._adapters_v1.utils import maybe_rekey_v0_key as _maybe_rekey_v0_key
from tamm._helpers import case_insensitive_lookup

_logger = _logging.getLogger(__name__)

__all__ = [
    "AdaptedModelStateDictFormat",
    "init",
    "is_adapter_initialized",
    "has_adapter",
    "delete_adapter",
    "freeze_adapter",
    "unfreeze_adapter",
    "set_active_adapter",
    "unset_active_adapter",
    "merge_adapter",
    "get_all_adapter_ids",
    "get_all_active_adapter_ids",
    "map_state_dict",
    "load_state_dict",
]


class AdaptedModelStateDictFormat(str, _enum.Enum):
    """
    An enum for controlling the :meth:`state_dict` behavior of :py:class:`AdaptedModel`

    FULL - load/get state_dict corresponding to the wrapped model and all adapters
    BASE - load/get state_dict corresponding to only the wrapped model
    ADAPTER - load/get state_dict corresponding to only the adapters
    """

    FULL = "FULL"
    BASE = "BASE"
    ADAPTER = "ADAPTER"

    @classmethod
    def _missing_(cls, value):
        return case_insensitive_lookup(cls, value)


def init(model: _nn.Module) -> _nn.Module:
    """
    Prepares model for adaptation.

    Args:
        model (:py:class:`_nn.Module`): Top level model which is adapted.

    Returns:
        :py:class:`_nn.Module` where adaptable layers are wrapped with :py:class:`_AdaptedLayer`
        wrappers.
    """
    adapted_map = {}
    children_to_wrap = []

    for parent in model.modules():
        for name, child in _helpers.get_all_named_children(parent):
            if _is_adaptable_layer(child) and not isinstance(child, _AdaptedLayer):
                if child not in adapted_map:
                    adapted_map[child] = _AdaptedLayer(child)
                children_to_wrap.append((parent, name, adapted_map[child]))

    for parent, name, new_child in children_to_wrap:
        parent.add_module(name, new_child)

    if _is_adaptable_layer(model):
        model = _AdaptedLayer(model)

    return model


def _verify_no_adapters(module: _AdaptedLayer, force: bool = False):
    if len(module.adapters) > 0:
        adapter_ids = list(module.adapters.keys())
        if force:
            _warnings.warn(
                f"Removing adapters with ids: {adapter_ids} from layer. If this is not "
                f"the intended outcome, consider merging the adapters before removing them."
            )
        else:
            raise RuntimeError(
                f"Trying to remove adapters with ids: {adapter_ids} from layer. "
                f"Merge them or pass force=True to force remove."
            )


def uninit(model: _nn.Module, force: bool = False) -> _nn.Module:
    """
    Removes adapters and adapted layer from the model.

    Args:
        model (:py:class:`_nn.Module`): Top level adapted model.
        force (:obj:`bool`): If force is ``True``, adapted layers are
            removed even if some layers have unmerged adapters, otherwise
            unmerged adapters lead to ``RuntimeError``.

    Returns:
    """
    adapted_map = {}
    children_to_unwrap = []

    for parent in model.modules():
        for name, child in _helpers.get_all_named_children(parent):
            if isinstance(child, _AdaptedLayer):
                _verify_no_adapters(child, force=force)
                if child not in adapted_map:
                    adapted_map[child] = child.wrapped
                children_to_unwrap.append((parent, name, adapted_map[child]))

    for parent, name, new_child in children_to_unwrap:
        parent.add_module(name, new_child)

    if isinstance(model, _AdaptedLayer):
        _verify_no_adapters(model, force=force)
        model = model.wrapped

    return model


def is_adapter_initialized(model: _nn.Module) -> bool:
    """
    Returns ``True`` if the model has been prepared for adaptation by wrapping with :py:func:`init`.
    """
    for _, module in model.named_modules(remove_duplicate=True):
        if isinstance(module, _AdaptedLayer):
            return True
    return False


def has_adapter(model: _nn.Module, adapter_id: str):
    """
    Returns ``True`` if the model has an adapter with ``adapter_id``.

    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters.
        adapter_id (:obj:`str`): The unique identifier for the adapter(s)
    """
    for _, module in model.named_modules(remove_duplicate=True):
        if isinstance(module, _AdaptedLayer) and module.has_adapter(adapter_id):
            return True
    return False


def _get_adapted_layers(model):
    return [layer for layer in model.modules() if isinstance(layer, _AdaptedLayer)]


def _apply_to_submodules(
    model: _nn.Module, apply_fn: str, adapter_id: _Optional[str] = None, **kwargs: _Any
):
    """
    Apply the function ``apply_fn`` to all submodules which are wrapped with
    :py:class:`_AdaptedLayer` wrappers.
    """
    if not has_adapter(model, adapter_id):
        raise RuntimeError(f"Model does not have adapter: {adapter_id}")

    for mod in _get_adapted_layers(model):
        if adapter_id and mod.has_adapter(adapter_id):
            getattr(mod, apply_fn)(adapter_id, **kwargs)
        elif adapter_id is None:
            getattr(mod, apply_fn)()


def delete_adapter(model: _nn.Module, adapter_id: str):
    """
    Deletes adapters identified by ``adapter_id``, if they exist.

    Args:
        model (:py:class:`_nn.Module`): Top level model, which has been prepared for
            adaptation with a call to :py:func:`init`, from which adapters are to be deleted.
        adapter_id (:obj:`str`): The unique identifier for the adapter(s) being
          deleted.
    """
    _apply_to_submodules(model, "delete_adapter", adapter_id)


def freeze_adapter(model: _nn.Module, adapter_id: str):
    """
    Freeze the parameters of the adapter(s) identified by ``adapter_id``.

    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters to be frozen.
        adapter_id (:obj:`str`): The unique identifier for the
          adapter(s) being frozen.
    """
    _apply_to_submodules(model, "freeze_adapter", adapter_id)


def unfreeze_adapter(model: _nn.Module, adapter_id: str):
    """
    Make the parameters of the adapter(s) identified by ``adapter_id`` trainable.

    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters to be un-frozen.
        adapter_id (:obj:`str`): The unique identifier for the adapter(s)
          being made trainable.
    """
    _apply_to_submodules(model, "unfreeze_adapter", adapter_id)


def set_active_adapter(model: _nn.Module, adapter_id: str):
    """
    Set the adapter already added to the layers as active adapter.
    All other adapters besides these are inactive and won't be used during
    model's forward pass.

    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters to activated.
        adapter_id (:obj:`str`): The unique identifier for the adapter(s)
          to be activated.
    """
    # call layer API with replace False so that we can catch instances where multiple
    # adapters are being activated for the same layer.
    _apply_to_submodules(model, "set_active_adapter", adapter_id, replace=False)


def unset_active_adapter(model: _nn.Module, adapter_id: str):
    """
    Unset the active adapter with the given adapter_id from the adapted layers.

    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters to activated.
        adapter_id (:obj:`str`): The unique identifier for the adapter(s)
          to be de-activated.
    """
    for mod in _get_adapted_layers(model):
        if mod.has_adapter(adapter_id):
            getattr(mod, "unset_active_adapter")()


def merge_adapter(model: _nn.Module, adapter_id: str):
    """
    Merge weights of adapter(s) with adapter_id into the model,
    if the adapter supports merging.

    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters to merged.
        adapter_id (:obj:`str`): The unique identifier for the adapter(s)
          to be merged.
    """
    _apply_to_submodules(model, "_merge_adapter", adapter_id)


def get_all_adapter_ids(model: _nn.Module) -> _List[str]:
    """
    Returns list of adapter ids of all unique adapters added to the model.

    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters.
    """
    adapter_ids = set()
    for _, module in model.named_modules(remove_duplicate=True):
        if isinstance(module, _AdaptedLayer):
            adapter_ids.update(module.adapters.keys())
    return sorted(list(adapter_ids))


def get_all_active_adapter_ids(model: _nn.Module) -> _List[str]:
    """
    Returns list of adapter ids of all unique adapters added to the model,
    which are active.

    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters.
    """
    adapter_ids = set()
    for _, module in model.named_modules(remove_duplicate=True):
        if isinstance(module, _AdaptedLayer) and module.active_adapter_id is not None:
            adapter_ids.add(module.active_adapter_id)
    return sorted(list(adapter_ids))


# pylint: disable=protected-access
def setup_adapter_extra_input_forwarding(
    model: _nn.Module, adapter_id: str, adapter_extra_inputs: _List[str]
):
    """
    Uses :py:class:`SharedContext` to enable extra inputs passed to this
    :py:class:`AdaptedModel` to be forwarded to adapters which consume
    these extra inputs.

    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters.
        adapter_id (:obj:`str`): The unique identifier for the adapter
            for which input forwarding is established
        adapter_extra_inputs (:obj:`list` of :obj:`str`): A
            list of names of extra inputs passed to this :py:class:`AdaptedModel`
            which are forwarded to adapters which consume these extra inputs.
    """
    # remove any handles registered for previous active adapter
    if hasattr(model, "_adapter_extra_input_forwarding_handles"):
        for handle in model._adapter_input_forwarding_handles:
            handle.remove()

    context = _SharedContext()

    creation_handle = context.register_context_creation_hook(model)
    population_handle = context.register_context_population_hook(
        model, context.get_input_forwarding_hook(*adapter_extra_inputs)
    )
    deletion_handle = context.register_context_deletion_hook(model)
    model._adapter_extra_input_forwarding_handles = [
        creation_handle,
        population_handle,
        deletion_handle,
    ]

    for mod in _get_adapted_layers(model):
        if mod.has_adapter(adapter_id):
            adapter = mod.get_adapter(adapter_id)
            if isinstance(adapter, _AdapterWithExtraInputs):
                context.register_context_retrieval_hook(
                    adapter, *adapter.extra_input_names
                )


def _normalize_wrapped_key(key: str):
    ignore_key_substrs = [
        "._checkpoint_wrapped_module",
        "_fsdp_wrapped_module.",
        "._fsdp_wrapped_module",
        "_fsdp_module.",
    ]
    for ignore_substr in ignore_key_substrs:
        if ignore_substr in key:
            key = key.replace(ignore_substr, "")
    return key


def _create_base_module_state_dict(model: _nn.Module) -> _Dict[str, _Any]:
    """
    Helper function for creating state_dict which only contains
    state corresponding to the wrapped model.
    """
    state_dict = model.state_dict()
    for module_name, module in model.named_modules():
        if not isinstance(module, _AdaptedLayer):
            continue
        for key in module.state_dict():
            original_key = f"{module_name}.{key}"
            original_key = _normalize_wrapped_key(original_key)
            # when this is called in FSDP.state_dict_type context, non zero rank
            # processes might have empty dictionaries, hence we need to check for presence of
            # original_key in state_dict
            if key.startswith("adapters") and original_key in state_dict:
                state_dict.pop(original_key)
            else:
                unwrapped_key = module_name + key[len("wrapped") :]
                unwrapped_key = _normalize_wrapped_key(unwrapped_key)
                if original_key in state_dict:
                    state_dict[unwrapped_key] = state_dict.pop(original_key)
    return state_dict


def _create_adapter_only_state_dict(
    model: _nn.Module,
    adapter_ids: _Optional[_List[str]] = None,
) -> _Dict[str, _Any]:
    """
    Helper function for creating state_dict which only contains
    state corresponding to the adapters identified by ``adapter_ids``.
    """

    # remove all keys not corresponding to state of adapter_ids
    keys_to_retain = set()
    if adapter_ids is None:
        adapter_ids = get_all_adapter_ids(model)

    for module_name, module in model.named_modules():
        if not isinstance(module, _AdaptedLayer):
            continue

        for adapter_id in adapter_ids:
            if not module.has_adapter(adapter_id):
                continue
            for key, _ in module.state_dict().items():
                if key.startswith(f"adapters.{adapter_id}"):
                    original_key = f"{module_name}.{key}"
                    # state dict keys don't have checkpoint wrapping baked into them but
                    # module name does, so we need to account for that
                    original_key = _normalize_wrapped_key(original_key)
                    keys_to_retain.add(original_key)

    state_dict = model.state_dict()
    state_dict_keys = list(state_dict.keys())
    for key in state_dict_keys:
        if key not in keys_to_retain:
            state_dict.pop(key)

    return state_dict


def map_state_dict(
    model: _nn.Module,
    fmt=AdaptedModelStateDictFormat.FULL,
    adapter_ids: _Optional[_List[str]] = None,
):
    """
    Return a dictionary containing references to the whole state of the model.
    Provides functionality to retrieve state corresponding to
    just the wrapped model or the adapters.

    .. note::
       When :py:meth:`map_state_dict` is used for creating state dict for a model wrapped with
       :py:class:`torch.distributed.fsdp.FullyShardedDataParallel`, it should be called under the
       following context::

           from torch.distributed.fsdp import StateDictType

           save_policy = FullStateDictConfig(offload_to_cpu=True, rank0_only=True)
           with FullyShardedDataParallel.state_dict_type(
               model, StateDictType.FULL_STATE_DICT, save_policy
           ):
               adapter_sd = map_state_dict(model, fmt="adapter", adapter_ids=[...])
               base_sd = map_state_dict(model, fmt="base")

       It is also recommended to use ``use_orig_params=True`` with ``FullyShardedDataParallel``,
       otherwise, the generated state dicts from :py:meth:`map_state_dict` are not guaranteed to be
       correct.


    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters.
        fmt (:obj:`AdaptedModelStateDictFormat`): Format for state_dict.
            The following values are allowed: 1) "full" - the entire state_dict
            is returned including keys for state of wrapped model and adapters
            2) "base" - only state corresponding to wrapped model is returned
            3) "adapter" - only state corresponding to adapters with
            ``adapter_ids`` is returned. Defaults to "full".
        adapter_ids (:obj:`list` of `str`, optional): Id(s) of adapter(s) whose
            state_dict is being loaded. This argument is requires when fmt is
            "adapter".
    """
    fmt = _helpers.get_enum_member_from_name(AdaptedModelStateDictFormat, fmt)
    if fmt is AdaptedModelStateDictFormat.FULL:
        return model.state_dict()
    if fmt is AdaptedModelStateDictFormat.BASE:
        return _create_base_module_state_dict(model)
    # if fmt is AdaptedModelStateDictFormat.ADAPTER
    return _create_adapter_only_state_dict(model, adapter_ids)


def _partial_load_state_dict_helper(
    model: _nn.Module,
    args: _Any,
    kwargs: _Any,
    state_dict: _Dict[str, _Any],
    expected_keys: _Set[str],
    strict: bool,
) -> _IncompatibleKeys:
    """
    Helper function for loading a partial state_dict generated
    from :py:meth:`state_dict` which either only contains state
    corresponding to the wrapped model or to one of the adapters.
    """
    error_msgs = []

    kwargs["strict"] = False

    partial_unexpected_keys = [key for key in state_dict if key not in expected_keys]

    # remove unexpected keys from state_dict because they can be adapter or base model
    # keys which would load fine in the model, but we do not want them to be loaded
    filtered_state_dict = {
        key: state_dict[key] for key in expected_keys if key in state_dict
    }

    result = model.load_state_dict(filtered_state_dict, *args, **kwargs)

    missing_keys = set(result.missing_keys)
    missing_keys = {_normalize_wrapped_key(key) for key in missing_keys}
    partial_missing_keys = [key for key in missing_keys if key in expected_keys]

    if strict:
        if len(partial_unexpected_keys) > 0:
            unexpected_keys_str = ", ".join(f'"{k}"' for k in partial_unexpected_keys)
            error_msgs.insert(
                0, f"Unexpected key(s) in state_dict: {unexpected_keys_str}. "
            )
        if len(partial_missing_keys) > 0:
            missing_keys_str = ", ".join(f'"{k}"' for k in partial_missing_keys)
            error_msgs.insert(0, f"Missing key(s) in state_dict: {missing_keys_str}. ")

    if len(error_msgs) > 0:
        raise RuntimeError(
            "Error(s) in loading state_dict for {}:\n\t{}".format(
                model.__class__.__name__, "\n\t".join(error_msgs)
            )
        )
    return _IncompatibleKeys(partial_missing_keys, partial_unexpected_keys)


def load_state_dict(
    model: _nn.Module,
    state_dict,
    *args,
    fmt=AdaptedModelStateDictFormat.FULL,
    adapter_ids: _Optional[_List[str]] = None,
    **kwargs,
):
    """
    Copies parameters from ``state_dict`` into the model. This method supports
    loading state dicts with or without adapters.

    Args:
        model (:py:class:`_nn.Module`): Top level model with adapters.
        state_dict (:obj:`dict`): The state dict to load.
        args: Variable arguments to forward to the base :meth:`load_state_dict`
            method.
        kwargs: Keyword arguments to forward to the base :meth:`load_state_dict`
            method.
        fmt (:obj:`AdaptedModelStateDictFormat`): Format for state_dict being loaded
            The following values are allowed: 1) "full" - state_dict contains
            the entire state of wrapped model and adapters 2) "base" - state_dict
            contains only state corresponding to wrapped model 3) "adapter" -
            state_dict contains only state corresponding to adapters with
            ``adapter_ids``. Defaults to "full". Defaults to "base".
        adapter_ids (:obj:`list` of `str`, optional): Id(s) of adapter(s) whose
            state_dict is being loaded. This argument is required when
            fmt is "adapter".
    """
    strict = kwargs.get("strict", True)
    fmt = _helpers.get_enum_member_from_name(AdaptedModelStateDictFormat, fmt)
    current_state_dict = map_state_dict(model, fmt=fmt, adapter_ids=adapter_ids)

    if _is_v0_lora_state_dict(state_dict):
        _warnings.warn(
            "Found state dict in v0 format. v0 adapters are deprecated, and "
            "support will be removed in a future tamm version.",
            category=DeprecationWarning,
        )
        state_dict = _maybe_rekey_v0_key(
            v1_state_dict=current_state_dict, v0_state_dict=state_dict
        )
    state_dict = _maybe_rekey_state_dict(model, state_dict)

    if fmt is AdaptedModelStateDictFormat.FULL:
        result = model.load_state_dict(state_dict, *args, **kwargs)
    elif fmt is AdaptedModelStateDictFormat.BASE:
        base_state_dict = _maybe_rekey_state_dict(model, current_state_dict)
        result = _partial_load_state_dict_helper(
            model, args, kwargs, state_dict, set(base_state_dict), strict
        )
    else:  # elif fmt is AdaptedModelStateDictFormat.ADAPTER
        if adapter_ids is None:
            raise ValueError(
                "adapter_id cannot be None when state_dict() is in ADAPTER mode"
            )
        for adapter_id in adapter_ids:
            if not has_adapter(model, adapter_id):
                raise RuntimeError(
                    f"Cannot find adapter with id: {adapter_ids}. state_dict "
                    f"cannot be loaded."
                )

        result = _partial_load_state_dict_helper(
            model, args, kwargs, state_dict, set(current_state_dict), strict
        )

    return result
